##############################################################################
#    TCL Developer Studio
#
#    Copyright (C) 1999  Alexey Kakunin
#    small@star.spb.ru
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
##############################################################################

namespace eval editorWindows {
    
    namespace export create openFile selectFile applyChanges isChanged deleteFile
    namespace export copy cut paste delete selectAll undo redo
    namespace export gotoMark gotoProc find findNext setCursor replace replaceAll
    namespace export enableHL disableHL onTabSize onFontChange
    namespace export colorizeFile
    namespace export formatSelection tabifySelection untabifySelection uppercaseSelection 
    namespace export lowercaseSelection stripSelection indentSelection unindentSelection commentSelection
    variable This
    variable TxtWidget ""
    variable Text ""
    variable UndoID
    
    # editorWindows::create --
    #
    #        Create editors frame of TCL DevStudio
    #
    #        Arguments:
    #            this        frame for storing editors windows
    #
    #        Results: (no results)
    proc create {this} {
        global text
        
        variable This
        
        frame $this.f -bg grey50 -relief sunken
        pack $this.f -expand 1 -fill both
        
        set This $this.f
        return
    }
    
    # editorWindows::openFile --
    #
    #        Open file. Add file to editted files.
    #        Arguments:
    #                fileName - name of file for loading
    #
    #        Results: (no results)
    proc openFile {fileName} {
        global tclDevData
        variable Text
        variable This
        
        if {$Text == ""} {
            # create text widget
            CreateTxtWidget
        }
        
        # Read Marks
        ReadMarks $fileName
        
        # set text to the widget
        selectFile $fileName
        
        return
    }
    
    # editorWindows::CreateTxtWidget --
    #
    #        Create Text Widget
    #        Arguments: (none)
    #
    #        Results: (none)
    proc CreateTxtWidget {} {
        global tcl_platform
        global tclDevData
        
        variable This
        variable TxtWidget
        variable Text
        variable UndoID
        
        if {$Text != ""} {
            error "Internal Error"
        }
        
        $This configure -bg grey
        set Text [tixScrolledText $This.text -borderwidth 10]
        
        set TxtWidget [$Text subwidget text]
        set UndoID [new textUndoer $TxtWidget]        
        # create font for editor window
        #    if {$tcl_platform(platform) == "unix"} {
            #        set size 12
            #    } else {
            #        set size 10
            #    }
        
        set tabSize [expr {$tclDevData(options,tabSize)*
            [font measure $tclDevData(options,fonts,editorFont) -displayof $TxtWidget " "]}]
        
        $TxtWidget configure -wrap none -font $tclDevData(options,fonts,editorFont) -tabs [list $tabSize]
        
        ConfigureTags
        
        pack $Text -expand 1 -fill both -padx 5 -pady 5
        
        # create bindings
        bind $TxtWidget <Tab> "editorWindows::OnTabPress; break"
        
        bind $TxtWidget <Control-Insert> "editorWindows::copy; break"
        bind $TxtWidget <Control-Delete> "editorWindows::cut; break"
        bind $TxtWidget <Shift-Insert>   "editorWindows::paste; break"
        
        bind $TxtWidget <Control-a> "editorWindows::selectAll; break"
        
        bind $TxtWidget <KeyRelease-Return> "editorWindows::IndentCurLine;editorWindows::OnKeyRelease"
        bind $TxtWidget <KeyRelease-braceright> "editorWindows::IndentCurLine;editorWindows::OnKeyRelease"
        
        bind $TxtWidget <KeyRelease-space> "editorWindows::OnSpaceRelease;editorWindows::OnKeyRelease"
        bind $TxtWidget <KeyRelease-braceleft> "editorWindows::OnLeftBraceRelease;editorWindows::OnKeyRelease"
        bind $TxtWidget <KeyRelease-parenleft> "editorWindows::OnLeftParenRelease;editorWindows::OnKeyRelease"
        bind $TxtWidget <KeyRelease-bracketleft> "editorWindows::OnLeftBracketRelease;editorWindows::OnKeyRelease"
        bind $TxtWidget <KeyRelease-quotedbl> "editorWindows::OnQuoteDblRelease;editorWindows::OnKeyRelease"
        
        bind $TxtWidget <KeyRelease> editorWindows::OnKeyRelease
        bind $TxtWidget <ButtonRelease> editorWindows::OnMouseRelease
        
        bind $TxtWidget <Control-n> "mainFrame::onFileNew; break"
        bind $TxtWidget <Control-o> "mainFrame::onFileOpen; break"
        bind $TxtWidget <Control-s> "mainFrame::onFileSave; break"
        bind $TxtWidget <Control-p> "mainFrame::onFilePrint; break"
        bind $TxtWidget <Control-x> "editorWindows::cut; break"
        bind $TxtWidget <Control-c> "editorWindows::copy; break"
        bind $TxtWidget <Control-v> "editorWindows::paste; break"
        bind $TxtWidget <Control-f> "editorWindows::OnFind; break"
        bind $TxtWidget <Control-l> "editorWindows::findNext; break"
        bind $TxtWidget <Control-h> "editorWindows::OnReplace; break"
        bind $TxtWidget <Control-g> "mainFrame::goTo; break"
        bind $TxtWidget <Control-z> "editorWindows::undo;editorWindows::OnKeyRelease"    
        bind $TxtWidget <Control-y> "editorWindows::redo;editorWindows::OnKeyRelease"    
        
        bind $TxtWidget <Control-F> "editorWindows::formatSelection"
        bind $TxtWidget <Control-C> "editorWindows::commentSelection"
        bind $TxtWidget <Control-I> "editorWindows::indentSelection"
        bind $TxtWidget <Control-U> "editorWindows::unindentSelection"
        
        #enable menus and buttons
        mainFrame::onCreateText
        
        return
    }
    
    # delete text widget
    proc DeleteTxtWidget {} {
        
        variable This
        variable Text
        variable TxtWidget
        variable UndoID
        
        if {$Text == ""} {
            error "Internal Error"
        }
        
        ::delete textUndoer $UndoID
        
        tixDestroy $Text
        set Text ""
        set TxtWidget ""
        
        $This configure -bg grey50
        
        #disable menus and buttons
        mainFrame::onDestroyText
        
        return
    }
    
    # editorWindows::selectFile --
    #
    #        Set file as current file
    #        Arguments: fileName
    #
    #        Results: (none)
    proc selectFile {fileName} {
        global tclDevData
        variable TxtWidget
        
        if {$fileName != $tclDevData(curFile)} {
            
            # this is not used now.       
            #        if {$tclDevData(curFile) != ""} {
                #            projectWindow::closeFile $tclDevData(curFile)
                #        }
            
            # store information about current file
            applyChanges
            
            # set new text
            SetFile $fileName
            
            # set focus to the editor window
            focus $TxtWidget
        }
    }
    
    # set active file
    proc SetFile {fileName} {
        global tclDevData
        variable TxtWidget
        
        #clear text widget
        $TxtWidget delete 0.0 end
        
        set tclDevData(curFile) $fileName
        $TxtWidget insert 1.0 $tclDevData(files,$tclDevData(curFile),text)
        $TxtWidget mark set insert $tclDevData(files,$tclDevData(curFile),cursor)
        
        if {$tclDevData(files,$tclDevData(curFile),selFirst) != ""} {
            $TxtWidget tag add sel \
            $tclDevData(files,$tclDevData(curFile),selFirst) \
            $tclDevData(files,$tclDevData(curFile),selLast)
        }
        
        applyMarks
        projectWindow::applyMarks $tclDevData(curFile)
        
        tixWidgetDoWhenIdle editorWindows::colorize $TxtWidget
        
        $TxtWidget see insert
        ReadCursor
        
        if {$tclDevData(files,$tclDevData(curFile),anchor) != ""} {
            $TxtWidget mark set anchor $tclDevData(files,$tclDevData(curFile),anchor)
        }
        
        return
    }
    
    # read changes from text widget to the internal buffer
    proc applyChanges {} {
        global tclDevData
        variable TxtWidget
        
        if {$tclDevData(curFile) != ""} {
            set text [$TxtWidget get 1.0 "end -1c"]
            if {$text != $tclDevData(files,$tclDevData(curFile),text)} {
                set tclDevData(files,$tclDevData(curFile),isChanged) 1
            }
            
            set tclDevData(files,$tclDevData(curFile),text) [$TxtWidget get 1.0 "end -1c"]
            set tclDevData(files,$tclDevData(curFile),cursor) [$TxtWidget index insert]
            
            if {[catch {$TxtWidget index anchor}]} {
                set tclDevData(files,$tclDevData(curFile),anchor) ""
            } else {
                set tclDevData(files,$tclDevData(curFile),anchor) [$TxtWidget index anchor]
            }
            
            if {[catch {$TxtWidget index sel.first}]} {
                set tclDevData(files,$tclDevData(curFile),selFirst) ""
                set tclDevData(files,$tclDevData(curFile),selLast) ""
            } else {
                set tclDevData(files,$tclDevData(curFile),selFirst) [$TxtWidget index sel.first]
                set tclDevData(files,$tclDevData(curFile),selLast) [$TxtWidget index sel.last]
            }
            
            ReadMarks $tclDevData(curFile)
            applyMarks
            
            # read information about comments and keywords
            set tclDevData(files,$tclDevData(curFile),comments) [$TxtWidget tag ranges comment]
            set tclDevData(files,$tclDevData(curFile),keywords) [$TxtWidget tag ranges keyword]
            
            projectWindow::applyMarks $tclDevData(curFile)
        }
        
        return   
    }
    
    #is file Changed?
    proc fileChanged {fileName} {
        global tclDevData
        
        variable TxtWidget
        
        if {$tclDevData(files,$fileName,isChanged)} {
            return 1
        }
        
        if {$fileName == $tclDevData(curFile)} {
            if {$tclDevData(files,$tclDevData(curFile),text) != [$TxtWidget get 1.0 "end -1c"]} {
                return 1
            } 
        }
        
        return 0
    }
    
    # reaction on closing file.
    proc deleteFile {fileName {newFile ""}} {
        global tclDevData
        variable TxtWidget
        
        if {$newFile == ""} {
            #delete Text Widget
            DeleteTxtWidget
            
            #reset curFile variable
            set tclDevData(curFile) ""
        } else {
            # set new File
            SetFile $newFile
            
            # set focus to text widget
            focus $TxtWidget
        }
        
        return
    }
    
    # edit-copy
    proc copy {} {
        variable TxtWidget
        
        if {[catch {$TxtWidget index sel.first}]} {
            return
        }
        
        set lineStart [lindex [split [$TxtWidget index sel.first] "."] 0]
        set lineEnd [lindex [split [$TxtWidget index sel.last] "."] 0]
        
        tk_textCopy $TxtWidget
        
        ReadCursor
        ColorizeLines $lineStart $lineEnd
        
        return
    }
    
    # edit-cut
    proc cut {} {
        variable TxtWidget
        
        if {$TxtWidget == ""} {
            return
        }
        
        tk_textCut $TxtWidget
        
        applyChanges
        ReadCursor
        
        set lineNum [lindex [split [$TxtWidget index insert] "."] 0]
        
        ColorizeLines $lineNum $lineNum
        
        return
    }
    
    # edit-paste
    proc paste {} {
        global tcl_platform
        variable TxtWidget
        
        if {$TxtWidget == ""} {
            return
        }
        
        if {"$tcl_platform(platform)" == "unix"} {
            catch { $TxtWidget delete sel.first sel.last }
        }
        
        set lineStart [lindex [split [$TxtWidget index insert] "."] 0]
        
        tk_textPaste $TxtWidget
        $TxtWidget see insert
        
        applyChanges
        ReadCursor
        
        set lineEnd [lindex [split [$TxtWidget index insert] "."] 0]
        ColorizeLines $lineStart $lineEnd
        
        return
    }
    
    # edit-delete
    proc delete {} {
        global tcl_platform
        variable TxtWidget
        
        if {$TxtWidget == ""} {
            return
        }
        
        catch { $TxtWidget delete sel.first sel.last }
        
        applyChanges
        ReadCursor
        set lineNum [lindex [split [$TxtWidget index insert] "."] 0]
        
        ColorizeLines $lineNum $lineNum
        
        return
    }
    
    proc selectAll {} {
        variable TxtWidget
        
        if {$TxtWidget == ""} {
            return
        }
        
        $TxtWidget tag add sel 0.0 end
        $TxtWidget mark set  anchor insert
    }
    
    # set cursor to the function
    proc gotoMark { markName } {    
        variable TxtWidget
        
        $TxtWidget mark set insert $markName
        $TxtWidget see insert
        focus $TxtWidget
        ReadCursor 0
    }
    
    proc gotoProc {procName} {
        variable TxtWidget
        
        set result [$TxtWidget search -regexp -- "^( |\t)*proc( |\t)+$procName" insert]
        
        if {$result != ""} {
            $TxtWidget mark set insert $result
            $TxtWidget see insert
            focus $TxtWidget
            ReadCursor 0
        }
        
        return
    }
    
    # parse file and create proc file
    proc ReadMarks { fileName } {
        global tclDevData
        
        # clear all marks in this file
        foreach name [array names tclDevData files,$tclDevData(curFile),marks,] {
            unset tclDevData($name)
        }
        
        set tclDevData(files,$fileName,marks) {}
        
        set lineNum 1
        #gets lines from text
        set lines [split $tclDevData(files,$fileName,text) '\n']
        
        # for each line
        foreach line $lines {
            # is it procedure definition
            if { [regexp "^( |\t)*proc( |\t)+(.*)" $line m a b rest] } {
                 if { [regexp "^\{(\[^\}\]*)" $rest m name] == 0 } {
                     set n [string first " " $rest]
                     set name [string range $rest 0 [expr $n - 1]]                
                 }            
                
                set markName $name
                append markName $lineNum
                
                lappend tclDevData(files,$fileName,marks) $markName
                set tclDevData(files,$fileName,marks,$markName,lineNum) $lineNum
                set tclDevData(files,$fileName,marks,$markName,name) $name
            }
            
            incr lineNum
        }
        
        return
    }
    
    # set procedure marks into the text widget
    proc applyMarks { } {
        global tclDevData
        variable TxtWidget
        
        # delete old marks
        foreach markName [$TxtWidget mark names] {
            if {$markName != "insert"} {
                $TxtWidget mark unset $markName
            }
        }
        
        # set all marks
        foreach markName $tclDevData(files,$tclDevData(curFile),marks) {
            $TxtWidget mark set $markName \
            $tclDevData(files,$tclDevData(curFile),marks,$markName,lineNum).0
        }
    }
    
    proc IndentCurLine {} {
        variable TxtWidget
        IndentLine [lindex [split [$TxtWidget index insert] "."] 0]
    }
    
    proc IndentLine {lineNum} {
        variable TxtWidget
        global tclDevData
        
        if {!$tclDevData(options,useIndent)} {
            return
        }
        
        if {$lineNum > 1} {
            # get current text
            set curText [$TxtWidget get $lineNum.0 "$lineNum.0 lineend"]
            
            #get text of prev line
            set result [GetOpenPair "\}" "$lineNum.0"]
            if {$result != ""} {
                set prevLineNum [lindex [split $result "."] 0]
            } else  {
                set prevLineNum [expr {$lineNum - 1}]
            }
            
            set prevText [$TxtWidget get $prevLineNum.0 "$prevLineNum.0 lineend"]
            
            #count first spaces in current line
            set spaces ""
            regexp "^\ *" $curText spaces
            
            #count first spaces in prev line
            set prevSpaces ""
            regexp "^(\ |\t)*" $prevText prevSpaces
            
            set len [string length $prevSpaces]
            set shouldBeSpaces 0
            
            for {set i 0} {$i < $len} {incr i} {
                if {[string index $prevSpaces $i] == "\t"} {
                    incr shouldBeSpaces $tclDevData(options,tabSize)
                } else  {
                    incr shouldBeSpaces
                }
            }
            
            if {$tclDevData(options,useSintaxIndent)} {
                #see last symbol in the prev String.
                set lastSymbol [string index $prevText [expr {[string length $prevText] - 1}]]
                
                # is it open brace?
                if {$lastSymbol == "\{"} {
                    incr shouldBeSpaces $tclDevData(options,indentSize)
                }
                
                # it first symbol in this string - close brace
                set a ""
                regexp "^\ *\}" $curText a
                
                if {$a != ""} {
                    # make unindent
                    if {$shouldBeSpaces >= $tclDevData(options,indentSize)} {
                        set shouldBeSpaces [expr {$shouldBeSpaces - $tclDevData(options,indentSize)}]
                    }
                }
            }
            
            set spaceNum [string length $spaces]
            if {$shouldBeSpaces > $spaceNum} {
                #insert spaces
                set deltaSpace [expr {$shouldBeSpaces - $spaceNum}]
                set incSpaces ""
                for {set i 0} {$i < $deltaSpace} {incr i} {
                    append incSpaces " "
                }
                $TxtWidget insert $lineNum.0 $incSpaces
            } elseif {$shouldBeSpaces < $spaceNum} {
                #delete spaces
                set deltaSpace [expr {$spaceNum - $shouldBeSpaces}]
                $TxtWidget delete $lineNum.0 $lineNum.$deltaSpace
            }
        }
        
        return
    }
    
    # change tab to spaces
    proc OnTabPress {} {
        variable TxtWidget
        global tclDevData
        
        if {$tclDevData(options,changeTabs)} {
            set spaces ""
            
            for {set i 0} {$i < $tclDevData(options,tabSize)} {incr i} {
                append spaces " "
            }
            
            #insert spaces
            $TxtWidget insert insert $spaces
        } else {
            #insert tab
            $TxtWidget insert insert "\t"
        }
    }
    
    # reaction on key releasing
    proc OnKeyRelease {} {
        variable TxtWidget
        #get lineNum
        set lineNum [lindex [split [$TxtWidget index insert] "."] 0]
        
        ReadCursor
        after idle editorWindows::ColorizeLine $lineNum
        after idle editorWindows::ColorizePair
    }
    
    #reaction on space release
    proc OnSpaceRelease {} {
        global tclDevData
        variable TxtWidget
        
        if {!$tclDevData(options,useTemplates) || !$tclDevData(options,useTemplatesForKeywords)} {
            return
        }
        
        set templateKeyword [GetTemplateKeyword [$TxtWidget get "insert linestart" "insert lineend"]]
        set curPos [$TxtWidget index insert]
        set lineNum [lindex [split $curPos "."] 0]
        
        switch $templateKeyword {
            "if" {
                $TxtWidget insert insert " \{\n\}"
                incr lineNum
                IndentLine $lineNum
                $TxtWidget mark set insert $curPos
            }
            
            "for" {
                $TxtWidget insert insert "\{\} \{\} \{\} \{\n\}"
                incr lineNum
                IndentLine $lineNum
                $TxtWidget mark set insert "$curPos +1ch"
            }
            
            "foreach" {
                $TxtWidget insert insert " \{\n\}"
                incr lineNum
                IndentLine $lineNum
                $TxtWidget mark set insert $curPos
            }
            
            "while" {
                $TxtWidget insert insert " \{\n\}"
                incr lineNum
                IndentLine $lineNum
                $TxtWidget mark set insert $curPos
            }
            
            "switch" {
                $TxtWidget insert insert " \{\n\}"
                incr lineNum
                IndentLine $lineNum
                $TxtWidget mark set insert $curPos
            } 
            
            "proc" {
                $TxtWidget insert insert " \{\} \{\n\}"
                incr lineNum
                IndentLine $lineNum
                $TxtWidget mark set insert $curPos
            }
            
            "else" {
                $TxtWidget insert insert " \{\n\n\}"
                ColorizeLine $lineNum
                incr lineNum
                IndentLine $lineNum
                $TxtWidget mark set insert "$lineNum.0 lineend"
                incr lineNum
                IndentLine $lineNum
            }
            
            "elseif" {
                $TxtWidget insert insert " \{\n\}"
                ColorizeLine $lineNum
                incr lineNum
                IndentLine $lineNum
                $TxtWidget mark set insert $curPos
            }
        }
        
        return 0
    }
    
    proc OnLeftBraceRelease {} {
        variable TxtWidget
        global tclDevData
        
        if {!$tclDevData(options,useTemplates) || !$tclDevData(options,useTemplatesForBrace)} {
            return
        }
        
        set curPos [$TxtWidget index insert]
        $TxtWidget insert insert "\}"
        $TxtWidget mark set insert $curPos
        
        return
    }
    
    proc OnLeftParenRelease {} {
        variable TxtWidget
        global tclDevData
        
        if {!$tclDevData(options,useTemplates) || !$tclDevData(options,useTemplatesForParen)} {
            return
        }
        
        set curPos [$TxtWidget index insert]
        $TxtWidget insert insert "\)"
        $TxtWidget mark set insert $curPos
        
        return
    }
    
    proc OnLeftBracketRelease {} {
        variable TxtWidget
        global tclDevData
        
        if {!$tclDevData(options,useTemplates) || !$tclDevData(options,useTemplatesForBracket)} {
            return
        }
        
        set curPos [$TxtWidget index insert]
        $TxtWidget insert insert "\]"
        $TxtWidget mark set insert $curPos
        
        return
    }
    
    proc OnQuoteDblRelease {} {
        variable TxtWidget
        global tclDevData
        
        if {!$tclDevData(options,useTemplates) || !$tclDevData(options,useTemplatesForQuoteDbl)} {
            return
        }
        
        set curPos [$TxtWidget index insert]
        $TxtWidget insert insert "\""
        $TxtWidget mark set insert $curPos
        
        return
    }
    
    
    # reasction on mouse button release
    proc OnMouseRelease {} {
        ReadCursor
        ColorizePair
    }
    
    # read information about cursor and set it to the global variables
    proc ReadCursor {{selectProc 1}} {
        variable TxtWidget
        global tclDevData
        
        set insertPos [split [$TxtWidget index insert] "."]
        set tclDevData(cursor,line) [lindex $insertPos 0]
        set tclDevData(cursor,pos) [expr {[lindex $insertPos 1] + 1}]
        
        if {$selectProc} {
            # -1 for skipping some apecial marks, such as "anchor"    
            set procName [$TxtWidget mark previous "insert -1 char"]
            
            if {$procName != ""} {
                projectWindow::selectProc $tclDevData(curFile) $procName
            }
        }
        
        return
    }
    
    proc find {text {options {}}} {
        variable TxtWidget
        
        set direction "forward"
        
        if {$text != "" && $TxtWidget != ""} {
            # get start position for searching
            if {[catch {$TxtWidget index sel.first}]} {
                set startPos insert
            } else {
                if {[lsearch $options -backward] != -1} {
                    set startPos [$TxtWidget index sel.first]
                    set direction "backward"
                } else {
                    set startPos [$TxtWidget index sel.last]
                }
            }
            
            set result ""
            if {$options != {}} {
                set result [eval "$TxtWidget search -count ::l $options -- \"$text\" $startPos"]
            } else  {
                set result [eval "$TxtWidget search -count ::l -- \"$text\" $startPos"]
            }
            
            if {$result != "" &&
                ($direction == "forward" && [$TxtWidget compare $startPos < $result] ||
                $direction == "backward" && [$TxtWidget compare $startPos > $result] ||
                [$TxtWidget compare $startPos == $result])} {
                $TxtWidget mark set insert "$result +$::l chars"
                $TxtWidget mark set anchor "$result"
                $TxtWidget tag remove sel 0.0 end
                $TxtWidget tag add sel "insert -$::l chars" insert
                $TxtWidget see insert
                focus $TxtWidget
                ReadCursor
            } else  {
                # put message into the status bar
                statusText::setText "\"$text\" not found."
            }
        }
        
        return
    }
    
    proc findNext {} {
        global tclDevData
        
        variable TxtWidget
        
        if {[llength $tclDevData(find)]} {
            find [lindex $tclDevData(find) 0] $tclDevData(find,lastOptions)
        }
        
        return
    }
    
    proc OnFind {} {
        variable TxtWidget
        
        #is has selection
        if {![catch {$TxtWidget index sel.first}]} {
            set text [$TxtWidget get [$TxtWidget index sel.first] [$TxtWidget index sel.last]]
        } else {
            # get current word
            set text [$TxtWidget get "insert wordstart" "insert wordend"]
        }
        
        mainFrame::find $text
    }
    
    proc OnReplace {} {
        variable TxtWidget
        
        # get current word
        set text [$TxtWidget get "insert wordstart" "insert wordend"]
        
        mainFrame::replace $text
    }
    
    proc enableHL {} {
        variable TxtWidget
        
        if {$TxtWidget != ""} {
            colorize
        }
        
        return
    }
    
    proc disableHL {} {
        variable TxtWidget
        
        if {$TxtWidget != ""} {
            # delete all tags
            $TxtWidget tag delete comment
            $TxtWidget tag delete keyword
            
            ConfigureTags
        }
        
        return
    }
    
    proc colorize {{windowName ""}} {
        variable TxtWidget
        global tclDevData
        
        if {!$tclDevData(options,useHL)} {
            return
        }
        
        if {$tclDevData(files,$tclDevData(curFile),comments) == {} ||
            $tclDevData(files,$tclDevData(curFile),keywords) == {}} {
            
            # get number of lines
            set lineEnd [lindex [split [$TxtWidget index end] "."] 0]
            
            ColorizeLines 1 $lineEnd
        } else {
            # delete tags
            $TxtWidget tag delete comment
            $TxtWidget tag delete keyword
            ConfigureTags
            
            # set tags
            eval $TxtWidget tag add comment $tclDevData(files,$tclDevData(curFile),comments)
            eval $TxtWidget tag add keyword $tclDevData(files,$tclDevData(curFile),keywords)
        }
    }
    
    proc ColorizeLines {lineStart lineEnd} {
        variable TxtWidget
        global tclDevData
        
        if {!$tclDevData(options,useHL)} {
            return
        }
        
        # delete all tags
        $TxtWidget tag remove comment "$lineStart.0" "$lineEnd.0 lineend"
        $TxtWidget tag remove keyword "$lineStart.0" "$lineEnd.0 lineend"
        
        for {set lineNum $lineStart} {$lineNum <= $lineEnd} {incr lineNum} {
            ColorizeLine $lineNum
        }
        
        return
    }
    
    proc ColorizeLine {lineNum} {
        variable TxtWidget
        global tclDevData
        
        if {!$tclDevData(options,useHL)} {
            return
        }
        
        #get line
        set line [$TxtWidget get $lineNum.0 "$lineNum.0 lineend"]
        
        set commentSpaces [IsComment $line]
        if {$commentSpaces != -1} {
            # this is comment
            #set comment font
            $TxtWidget tag add comment $lineNum.$commentSpaces "$lineNum.0 lineend"
            $TxtWidget tag remove keyword $lineNum.0 "$lineNum.0 lineend"
        } else {
            $TxtWidget tag remove comment $lineNum.0 "$lineNum.0 lineend"
            
            set l [GetKeywordCoord $line]
            
            if {$l != {} } {
                set nameStart [lindex $l 0]
                set nameEnd [lindex $l 1]
                
                $TxtWidget tag add keyword "$lineNum.$nameStart" "$lineNum.$nameEnd"
            } else {
                $TxtWidget tag remove keyword $lineNum.0 "$lineNum.0 lineend"
            }
        }
        
        return    
    }
    
    proc colorizeFile {fileName} {
        global tclDevData
        
        # clear old comments and keywords
        set tclDevData(files,$fileName,comments) {}
        set tclDevData(files,$fileName,keywords) {}
        
        if {!$tclDevData(options,useHL)} {
            return
        }
        
        # split file by lines
        set ltext [split $tclDevData(files,$fileName,text) "\n"]
        set lineNum 1
        
        foreach line $ltext {
            set commentSpaces [IsComment $line]
            if {$commentSpaces != -1} {
                # this is comment
                #set comment font
                lappend tclDevData(files,$fileName,comments) $lineNum.$commentSpaces "$lineNum.0 lineend"
            } else {
                set l [GetKeywordCoord $line]
                
                if {$l != {} } {
                    set nameStart [lindex $l 0]
                    set nameEnd [lindex $l 1]
                    
                    lappend tclDevData(files,$fileName,keywords) "$lineNum.$nameStart" "$lineNum.$nameEnd"
                }
            }
            
            incr lineNum
        }
        
        return
    }
    
    
    proc ConfigureTags {} {
        variable TxtWidget
        global tclDevData
        
        # blue is specially for Lapshin
        $TxtWidget tag configure comment -font $tclDevData(options,fonts,commentFont) -foreground blue
        $TxtWidget tag configure keyword -font $tclDevData(options,fonts,keywordFont)
        $TxtWidget tag configure pair -background red
        
        return    
    }

    # This proc returns number of characters before comment symbol
    # If string is not comment - return -1
    proc IsComment { line } {
        set a ""
        regexp "^( |\t)*\#" $line a
        
        if {$a != ""} {
            return [expr {[string length $a] - 1}]
        } else {
            return -1
        }
    }
    
    proc GetKeywordCoord {line} {
        global tclDevData
        
        set a ""
        regexp "^( |\t)*(\[a-z\]|\[A-Z\]|\[0-9\]|_|:|~|\\.|/)+" $line a
        
        if {$a != ""} {
            # gets name
            set b ""
            regexp "^( |\t)*" $line b
            set nameStart [string length $b]
            set nameEnd [string length $a]
            set name [string range $a [string length $b] end]
            
            # is it keyword?
            if {[lsearch $tclDevData(keywords) $name] != -1} {
                lappend l $nameStart
                lappend l $nameEnd
                return $l
            }
        } else {
            # check for else
            set a ""
            regexp "^( |\t)*\}( |\t)*else( |\t)*\{" $line a
            
            if {$a != ""} {
                set b ""
                regexp "^( |\t)*\}( |\t)*" $line b
                
                lappend l [string length $b]
                lappend l [expr {[string length $b] + 4}]
                
                return $l
            }
            
            # check for elseif
            set a ""
            regexp "^( |\t)*\}( |\t)*elseif( |\t)*\{" $line a
            
            if {$a != ""} {
                set b ""
                regexp "^( |\t)*\}( |\t)*" $line b
                
                lappend l [string length $b]
                lappend l [expr {[string length $b] + 6}]
                
                return $l
            }
        }
        
        return {}
    }
    
    proc GetTemplateKeyword { line } {
        global tclDevData
        
        set a ""
        regexp "^( |\t)*(\[a-z\]|\[A-Z\]|\[0-9\]|_|:|~|\\.|/)+ $" $line a
        
        if {$a != ""} {
            # gets name
            set b ""
            regexp "^( |\t)*" $line b
            set nameStart [string length $b]
            set nameEnd [string length $a]
            set name [string range $a [string length $b] end]
            
            #return name without last space
            return [string range $name 0 [expr {[string length $name] - 2}]]
        } else {
            # check for else
            set a ""
            regexp "^( |\t)*\}( |\t)*else $" $line a
            
            if {$a != ""} {
                return "else"
            }
            
            # check for elseif
            set a ""
            regexp "^( |\t)*\}( |\t)*elseif $" $line a
            
            if {$a != ""} {
                return "elseif"
            }
        }
        
        return ""
    }
    
    proc setCursor {lineNum pos} {
        variable TxtWidget
        
        $TxtWidget mark set insert $lineNum.$pos
        $TxtWidget see insert
        focus $TxtWidget
        ReadCursor
        
        return
    }
    
    proc replace {findStr replaceStr in options} {
        variable TxtWidget
        
        if {$findStr != "" && $TxtWidget != "" && 
            ($in == "file" || ![catch {$TxtWidget index sel.first}])} {
            
            #determine start pos
            if {$in == "file"} {
                set startPos insert
            } else {
                set startPos sel.first
            }
            
            if {$options != {}} {
                set result [eval "$TxtWidget search -count ::l $options -- \"$findStr\" $startPos"]
            } else  {
                set result [eval "$TxtWidget search -count ::l -- \"$findStr\" $startPos"]
            }
            
            if {$result != "" &&
                [$TxtWidget compare $startPos <= $result]} {
                # check , what it is current result
                if {$in != "file" && [lsearch [$TxtWidget tag names $result] sel] == -1} {
                    return
                }
                
                #delete old string
                $TxtWidget delete $result "$result +$::l chars"
                #insert new string
                $TxtWidget insert $result $replaceStr
                
                # set cursor
                $TxtWidget mark set insert "$result +[string length $replaceStr] chars"
                $TxtWidget see insert
                focus $TxtWidget
                ReadCursor
            }
        }
        
        return
    }
    
    proc replaceAll {findStr replaceStr in options} {
        variable TxtWidget
        
        set count 0
        
        if {$findStr != "" && $TxtWidget != "" && 
            ($in == "file" || ![catch {$TxtWidget index sel.first}])} {
            
            #determine start pos
            if {$in == "file"} {
                set startPos insert
            } else {
                set startPos sel.first
            }
            
            set result ""
            
            if {$options != {}} {
                set result [eval "$TxtWidget search -count ::l $options -- \"$findStr\" $startPos"]
            } else  {
                set result [eval "$TxtWidget search -count ::l -- \"$findStr\" $startPos"]
            }
            
            while {$result != "" &&
                [$TxtWidget compare $startPos <= $result]} {
                # check , what it is current result
                if {$in != "file" && [lsearch [$TxtWidget tag names $result] sel] == -1} {
                    return
                }
                
                #delete old string
                $TxtWidget delete $result "$result +$::l chars"
                #insert new string
                $TxtWidget insert $result $replaceStr
                
                # set cursor
                $TxtWidget mark set insert "$result +[string length $replaceStr] chars"
                
                incr count
                set result ""
                
                if {$options != {}} {
                    set result [eval "$TxtWidget search -count ::l $options -- \"$findStr\" $startPos"]
                } else  {
                    set result [eval "$TxtWidget search -count ::l -- \"$findStr\" $startPos"]
                }
            }
            
            $TxtWidget see insert
            focus $TxtWidget
            ReadCursor
        }
        
        statusText::setText "$count occurrence(s) have been replaces."
        
        return
    }
    
    #reaction on changing tab size
    proc onTabSize {} {
        variable TxtWidget
        global tclDevData
        
        if {$TxtWidget != ""} {
            set size [expr {$tclDevData(options,tabSize)*
                [font measure $tclDevData(options,fonts,editorFont) -displayof $TxtWidget " "]}]
            $TxtWidget configure -tabs [list $size]
        }
        
        return
    }
    
    # reaction on change font
    proc onFontChange {} {
        variable TxtWidget
        global tclDevData
        
        if {$TxtWidget != ""} {
            $TxtWidget configure -font $tclDevData(options,fonts,editorFont)
            ConfigureTags
        }
        
        return
    }
    
    proc GetOpenPair {symbol {index ""}} {
        variable TxtWidget
        
        if {$index == ""} {
            set index insert
        }
        
        set count 0
        set line [$TxtWidget get "$index linestart" $index]
        set lineNum [lindex [split [$TxtWidget index $index] "."] 0]
        
        set result [ProcessLineForOpenSymbol $line $symbol count]
        if {$result != ""} {
            #found
            return $lineNum.$result
        }
        
        incr lineNum -1
        while {$lineNum > 0} {
            set line [$TxtWidget get "$lineNum.0" "$lineNum.0 lineend"]
            set result [ProcessLineForOpenSymbol $line $symbol count]
            
            if {$result != ""} {
                return $lineNum.$result
            }
            
            incr lineNum -1
        }
        
        return ""
    }
    
    
    proc GetClosePair {symbol {index ""}} {
        variable TxtWidget
        
        if {$index == ""} {
            set index "insert +1ch"
        }
        
        set count 0
        set line [$TxtWidget get $index "$index lineend"]
        set lineNum [lindex [split [$TxtWidget index $index] "."] 0]
        set lineNums [lindex [split [$TxtWidget index end] "."] 0]
        
        set result [ProcessLineForCloseSymbol $line $symbol count]
        if {$result != ""} {
            #found
            incr result [lindex [split [$TxtWidget index $index] "."] 1]
            return $lineNum.$result
        }
        
        incr lineNum
        while {$lineNum < $lineNums} {
            set line [$TxtWidget get "$lineNum.0" "$lineNum.0 lineend"]
            set result [ProcessLineForCloseSymbol $line $symbol count]
            
            if {$result != ""} {
                return $lineNum.$result
            }
            
            incr lineNum
        }
        
        return ""
    }
    
    #process line for openSymbol
    proc ProcessLineForOpenSymbol {line symbol countName} {
        upvar $countName count
        
        switch $symbol {
            "\}" {
                set openSymbol "\{"
            }
            "\]" {
                set openSymbol "\["
            }
            "\)" {
                set openSymbol "\("
            }
        }
        
        #process line
        for {set i [expr {[string length $line] - 1}]} {$i >= 0} {incr i -1} {
            set curChar [string index $line $i]
            
            if {$curChar == $openSymbol} {
                # increment count
                if {[string index $line [expr {$i - 1}]] == "\\"} {
                    #skip it
                    incr i -1
                } else  {
                    incr count
                    if {$count > 0} {
                        return $i
                    }
                }
            } elseif {$curChar == $symbol } {
                # decrement count
                if {[string index $line [expr {$i - 1}]] == "\\"} {
                    #skip it
                    incr i -1
                } else  {
                    incr count -1
                }
            }
        }
        
        return ""
    }
    
    
    #process line for openSymbol
    proc ProcessLineForCloseSymbol {line symbol countName} {
        upvar $countName count
        
        switch $symbol {
            "\{" {
                set closeSymbol "\}"
            }
            "\[" {
                set closeSymbol "\]"
            }
            "\(" {
                set closeSymbol "\)"
            }
        }
        
        #process line
        set len [string length $line]
        for {set i 0} {$i < $len} {incr i } {
            set curChar [string index $line $i]
            
            if {$curChar == $closeSymbol} {
                # increment count
                incr count
                if {$count > 0} {
                    return $i
                }
            } elseif {$curChar == $symbol } {
                # decrement count
                incr count -1
            } elseif {$curChar == "\\"} {
                #skip next symbol
                incr i
            }
        }
        
        return ""
    }
    
    # count braces in line
    proc CountBraces {line count} {
        set len [string length $line]
        
        for {set i 0} {$i < $len} {incr i} {
            switch -- [string index $line $i] {
                "\{" {
                    # increment count
                    incr count
                }
                
                "\}" {
                    # decrement count
                    incr count -1
                }
                
                "\\" {
                    #skip next symbol
                    incr i
                }
            }
        }
        
        return $count
    }
    
    # colorize pair
    proc ColorizePair {} {
        variable TxtWidget
        
        $TxtWidget tag remove pair 0.0 end
        
        #get current char
        set curChar [$TxtWidget get insert "insert +1ch"]
        
        if {$curChar == "\}"} {
            set result [GetOpenPair "\}"]
            if {$result != ""} {
                $TxtWidget tag add pair $result "$result +1ch"
            }
        } elseif {$curChar == "\]"} {
            set result [GetOpenPair "\]"]
            if {$result != ""} {
                $TxtWidget tag add pair $result "$result +1ch"
            }
        } elseif {$curChar == "\)"} {
            set result [GetOpenPair "\)"]
            if {$result != ""} {
                $TxtWidget tag add pair $result "$result +1ch"
            }
        } elseif {$curChar == "\{"} {
            set result [GetClosePair "\{"]
            if {$result != ""} {
                $TxtWidget tag add pair $result "$result +1ch"
            }
        } elseif {$curChar == "\["} {
            set result [GetClosePair "\["]
            if {$result != ""} {
                $TxtWidget tag add pair $result "$result +1ch"
            }
        } elseif {$curChar == "\("} {
            set result [GetClosePair "\("]
            if {$result != ""} {
                $TxtWidget tag add pair $result "$result +1ch"
            }
        }
    }
    
    proc undo {} {
        variable UndoID
        textUndoer:undo $UndoID
        return
    }
    
    proc redo {} {
        variable UndoID
        textRedoer:redo $UndoID
        return
    }
    
    # a set of advanced functions
    proc formatSelection {} {
        variable TxtWidget
        global tclDevData
        
        #Gets start and end lines
        if {[catch {set selFirst [$TxtWidget index sel.first]}] ||
            [catch {set selLast [$TxtWidget index sel.last]}] ||
            [catch {set anchor [$TxtWidget index anchor]}]} {
            #nothing to do
            return
        }
        
        set startLine [lindex [split [$TxtWidget index sel.first] "."] 0]
        set endLine [lindex [split [$TxtWidget index sel.last] "."] 0]
        
        set count 0
        
        for {set lineNum 1} {$lineNum < $startLine} {incr lineNum} {
            set count [CountBraces [$TxtWidget get "$lineNum.0" "$lineNum.0 lineend"] $count]
        }
        
        for {} {$lineNum <= $endLine} {incr lineNum} {
            # indent line
            #count first spaces in current line
            set curText [$TxtWidget get "$lineNum.0" "$lineNum.0 lineend"]
            set spaces ""
            regexp "^\ *" $curText spaces
            
            set shouldBeSpaces [expr $count * $tclDevData(options,indentSize)]
            
            # it first symbol in this string - close brace
            set a ""
            regexp "^\ *\}" $curText a
            
            if {$a != ""} {
                # make unindent
                if {$shouldBeSpaces >= $tclDevData(options,indentSize)} {
                    set shouldBeSpaces [expr {$shouldBeSpaces - $tclDevData(options,indentSize)}]
                }
            }
            
            set spaceNum [string length $spaces]
            if {$shouldBeSpaces > $spaceNum} {
                #insert spaces
                set deltaSpace [expr {$shouldBeSpaces - $spaceNum}]
                set incSpaces ""
                for {set i 0} {$i < $deltaSpace} {incr i} {
                    append incSpaces " "
                }
                $TxtWidget insert $lineNum.0 $incSpaces
            } elseif {$shouldBeSpaces < $spaceNum} {
                #delete spaces
                set deltaSpace [expr {$spaceNum - $shouldBeSpaces}]
                $TxtWidget delete $lineNum.0 $lineNum.$deltaSpace
            }
            
            set count [CountBraces $curText $count]
        }
        
        return
    }
    
    proc tabifySelection {} {
        global tclDevData
        variable TxtWidget
        
        if {[catch {set selFirst [$TxtWidget index sel.first]}] ||
            [catch {set selLast [$TxtWidget index sel.last]}] ||
            [catch {set anchor [$TxtWidget index anchor]}]} {
            #nothing to do
            return
        }
        
        set insert [$TxtWidget index insert]
        set text [$TxtWidget get sel.first sel.last]
        
        #make spaces var
        for {set i 0} {$i < $tclDevData(options,tabSize)} {incr i} {
            append spaces " "
        }
        # convert spaces to tabs
        regsub -all -- $spaces $text "\t" newText
        set text $newText
        
        $TxtWidget delete sel.first sel.last
        $TxtWidget insert $selFirst $text
        
        # set selection
        $TxtWidget tag add sel $selFirst "$selFirst +[string length $text] chars"
        if {$anchor == $selLast} {
            $TxtWidget mark set anchor "$selFirst +[string length $text] chars"
        } else  {
            $TxtWidget mark set anchor $anchor
        }
        if {$insert == $selLast} {
            $TxtWidget mark set insert "$selFirst +[string length $text] chars"
        } else  {
            $TxtWidget mark set insert $insert
        }
        
        return
    }
    
    proc untabifySelection {} {
        global tclDevData
        variable TxtWidget
        if {[catch {set selFirst [$TxtWidget index sel.first]}] ||
            [catch {set selLast [$TxtWidget index sel.last]}] ||
            [catch {set anchor [$TxtWidget index anchor]}]} {
            #nothing to do
            return
        }
        
        set insert [$TxtWidget index insert]
        set text [$TxtWidget get sel.first sel.last]
        
        #make spaces var
        for {set i 0} {$i < $tclDevData(options,tabSize)} {incr i} {
            append spaces " "
        }
        # convert spaces to tabs
        regsub -all -- "\t" $text $spaces newText
        set text $newText
        
        $TxtWidget delete sel.first sel.last
        $TxtWidget insert $selFirst $text
        
        # set selection
        $TxtWidget tag add sel $selFirst "$selFirst +[string length $text] chars"
        if {$anchor == $selLast} {
            $TxtWidget mark set anchor "$selFirst +[string length $text] chars"
        } else  {
            $TxtWidget mark set anchor $anchor
        }
        if {$insert == $selLast} {
            $TxtWidget mark set insert "$selFirst +[string length $text] chars"
        } else  {
            $TxtWidget mark set insert $insert
        }
        return
    }
    
    proc uppercaseSelection {} {
        variable TxtWidget
        
        if {[catch {set selFirst [$TxtWidget index sel.first]}] ||
            [catch {set selLast [$TxtWidget index sel.last]}] ||
            [catch {set anchor [$TxtWidget index anchor]}]} {
            #nothing to do
            return
        }
        
        set insert [$TxtWidget index insert]
        set text [$TxtWidget get sel.first sel.last]
        
        set text [string toupper $text]
        
        $TxtWidget delete sel.first sel.last
        $TxtWidget insert $selFirst $text
        
        # set selection
        $TxtWidget tag add sel $selFirst $selLast
        $TxtWidget mark set anchor $anchor
        $TxtWidget mark set insert $insert
        
        return
    }
    
    proc lowercaseSelection {} {
        variable TxtWidget
        
        if {[catch {set selFirst [$TxtWidget index sel.first]}] ||
            [catch {set selLast [$TxtWidget index sel.last]}] ||
            [catch {set anchor [$TxtWidget index anchor]}]} {
            #nothing to do
            return
        }
        
        set insert [$TxtWidget index insert]
        set text [$TxtWidget get sel.first sel.last]
        
        set text [string tolower $text]
        
        $TxtWidget delete sel.first sel.last
        $TxtWidget insert $selFirst $text
        
        # set selection
        $TxtWidget tag add sel $selFirst $selLast
        $TxtWidget mark set anchor $anchor
        $TxtWidget mark set insert $insert
        
        return
    }
    
    proc stripSelection {} {
        variable TxtWidget
        global tclDevData
        
        #Gets start and end lines
        if {[catch {set selFirst [$TxtWidget index sel.first]}] ||
            [catch {set selLast [$TxtWidget index sel.last]}] ||
            [catch {set anchor [$TxtWidget index anchor]}]} {
            #nothing to do
            return
        }
        
        set startLine [lindex [split [$TxtWidget index sel.first] "."] 0]
        set endLine [lindex [split [$TxtWidget index sel.last] "."] 0]
        
        for {set lineNum $startLine} {$lineNum <= $endLine} {incr lineNum} {
            set curText [$TxtWidget get "$lineNum.0" "$lineNum.0 lineend"]

            set text [string trimright $curText]
            $TxtWidget delete "$lineNum.0" "$lineNum.0 lineend"
            $TxtWidget insert "$lineNum.0" $text
        }

        # set selection
        $TxtWidget tag add sel $selFirst $selLast
        $TxtWidget mark set anchor $anchor
        
        return
    }
    
    proc indentSelection {} {
        variable TxtWidget
        global tclDevData
        
        #Gets start and end lines
        if {[catch {set selFirst [$TxtWidget index sel.first]}] ||
            [catch {set selLast [$TxtWidget index sel.last]}] ||
            [catch {set anchor [$TxtWidget index anchor]}]} {
            #nothing to do
            return
        }
        
        set startLine [lindex [split [$TxtWidget index sel.first] "."] 0]
        set endLine [lindex [split [$TxtWidget index sel.last] "."] 0]
        if {$endLine == [lindex [split [$TxtWidget index end] "."] 0]} {
            #skip last line in widget
            incr endLine -1
        }
        
        for {set lineNum $startLine} {$lineNum <= $endLine} {incr lineNum} {
            set text " "
            append text [$TxtWidget get "$lineNum.0" "$lineNum.0 lineend"]
            
            $TxtWidget delete "$lineNum.0" "$lineNum.0 lineend"
            $TxtWidget insert "$lineNum.0" $text
        }
        # highlight
        ColorizeLines $startLine $endLine

        # set selection
        $TxtWidget tag add sel $selFirst $selLast
        $TxtWidget mark set anchor $anchor
        
        return
    }

    proc unindentSelection {} {
        variable TxtWidget
        global tclDevData
        
        #Gets start and end lines
        if {[catch {set selFirst [$TxtWidget index sel.first]}] ||
            [catch {set selLast [$TxtWidget index sel.last]}] ||
            [catch {set anchor [$TxtWidget index anchor]}]} {
            #nothing to do
            return
        }
        
        set startLine [lindex [split [$TxtWidget index sel.first] "."] 0]
        set endLine [lindex [split [$TxtWidget index sel.last] "."] 0]
        if {$endLine == [lindex [split [$TxtWidget index end] "."] 0]} {
            #skip last line in widget
            incr endLine -1
        }
        
        for {set lineNum $startLine} {$lineNum <= $endLine} {incr lineNum} {
            if {[$TxtWidget get "$lineNum.0" "$lineNum.0 +1 char"] == " "} {
                $TxtWidget delete "$lineNum.0" "$lineNum.0 +1 char"
            }
        }
        # highlight
        ColorizeLines $startLine $endLine

        # set selection
        $TxtWidget tag add sel $selFirst $selLast
        $TxtWidget mark set anchor $anchor
        
        return
    }
    
    proc commentSelection {} {
        variable TxtWidget
        global tclDevData
        
        #Gets start and end lines
        if {[catch {set selFirst [$TxtWidget index sel.first]}] ||
            [catch {set selLast [$TxtWidget index sel.last]}] ||
            [catch {set anchor [$TxtWidget index anchor]}]} {
            #nothing to do
            return
        }
 
        set insert [$TxtWidget index insert]
       
        set startLine [lindex [split [$TxtWidget index sel.first] "."] 0]
        set endLine [lindex [split [$TxtWidget index sel.last] "."] 0]
        
        if {$endLine == [lindex [split [$TxtWidget index end] "."] 0]} {
            #skip last line in widget
            incr endLine -1
        }
        
        for {set lineNum $startLine} {$lineNum <= $endLine} {incr lineNum} {
            set text "#"
            append text [$TxtWidget get "$lineNum.0" "$lineNum.0 lineend"]
            
            $TxtWidget delete "$lineNum.0" "$lineNum.0 lineend"
            $TxtWidget insert "$lineNum.0" $text
        }

        # highlight
        ColorizeLines $startLine $endLine
        
        # set selection
        $TxtWidget tag add sel $selFirst $selLast
        $TxtWidget mark set anchor $anchor
        $TxtWidget mark set insert $insert
        
        return
    }


}