##############################################################################
#    TCL Developer Studio
#
#    Copyright (C) 1999  Alexey Kakunin
#    small@star.spb.ru
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
##############################################################################

namespace eval mainFrame {

    namespace export create enable disable onCreateText onDestroyText find replace openTextFile
    namespace export onFileNew onFileOpen onFileSave goTo openFile onFilePrint
    namespace export getGeometry getProjectWinSize getOutputWinSize
    namespace export setProjectName refreshTree endRun selectMenu

    variable MainFrame
    variable MenuFrame
    variable ToolbarFrame
    variable StatusFrame
    variable PaneWindow1
    variable PaneWindow2
    variable RecentFilesMenu
    variable RecentProjectsMenu
    variable IsRunned 0
    variable IsDebugered 0

    # mainFrame::create --
    #
    #        Create main frame of TCL DevStudio
    #
    #        Arguments: (no arguments)
    #
    #        Results: (no results)
    proc create {} {
        global titlepane
        global text
        global tclDevData

        variable MainFrame
        variable StatusLabel
        variable PaneWindow1
        variable PaneWindow2

        # wait closing titlepane
        if {[winfo exist $titlepane]} {
            tkwait window $titlepane
        }

        # create frame
        set MainFrame .mainFrame
        toplevel $MainFrame -menu .mainFrame.menu
        
        # set geometry
        if {[info exist tclDevData(positions,mainFrame)] &&
             [string range $tclDevData(positions,mainFrame) 0 2] != "1x1"} {
            wm geometry $MainFrame $tclDevData(positions,mainFrame)
        }

        # set wm communication
        wm title $MainFrame "TCL Developer Studio"
        wm protocol $MainFrame WM_DELETE_WINDOW mainFrame::exit

        # Create Statusbar
        mainFrame::CreateStatusBar

        # Create menu for main frame
        mainFrame::CreateMenu
        # Create toolbar for main frame
        mainFrame::CreateToolBar

        # create paned windows
        tixPanedWindow $MainFrame.pane
        set PaneWindow1 $MainFrame.pane
        pack $MainFrame.pane -side top -expand 1 -fill both -padx 1 -pady 1

        if {[info exist tclDevData(positions,mainFrame)] &&
             [string range $tclDevData(positions,mainFrame) 0 2] != "1x1"} {
            set topWin [$MainFrame.pane add topWin -expand 1.0]
        } else {
            set topWin [$MainFrame.pane add topWin -size 200 -expand 1.0]
        }

        if {[info exist tclDevData(positions,outputWinSize)]} {
            set outWin [$MainFrame.pane add outWin -min 40 -size $tclDevData(positions,outputWinSize)]
        } else {
            set outWin [$MainFrame.pane add outWin -min 40 -size 120]
        }

        # Create new pane window in top window
        tixPanedWindow $topWin.pane -orientation horizontal
        set PaneWindow2 $topWin.pane
        pack $topWin.pane -side top -expand 1 -fill both

        #create project window and editors window
        if {[info exist tclDevData(positions,projectWinSize)]} {
            set projWin [$topWin.pane add project -size $tclDevData(positions,projectWinSize)]
        } else {
            set projWin [$topWin.pane add project -size 150]
        }

        if {[info exist tclDevData(positions,mainFrame)] &&
             [string range $tclDevData(positions,mainFrame) 0 2] != "1x1"} {
            set editsWin [$topWin.pane add editors -expand 1.0]
        } else {
            set editsWin [$topWin.pane add editors -size 200 -expand 1.0]
        }

        #Create sub-windows
        projectWindow::create $projWin
        editorWindows::create $editsWin
        outputWindow::create $outWin

        return
    }

    # mainFrame::exit --
    #
    #        Exit from TCL Developer Studio
    #
    #        Arguments: (no arguments)
    #
    #        Results: (no results)

    proc exit {} {
        global tclDevData
        # save options
        options::save

        # close all files
        if {[CloseAll]} {

            #delete images from menus.
            foreach imageName $tclDevData(menuImages) {
                image delete $imageName
            }

            ::exit
        }
    }

    proc CreateMenuItem {menuName menuText picName args} {
        global image
        global tclDevData

        # Icons in menus are disabled now.
        # This is very unstable feature
        set  label "-label \"$menuText\""

#        if {$picName == "" || !$tclDevData(options,useImagesInMenus)} {
#            set label "-label \"$menuText\""
#        } else  {
            # get menu background
#            set bg [$menuName cget -background]

            #create compound image
#            set cmpImage [image create compound -window $menuName -background $bg]
#            lappend tclDevData(menuImages) $cmpImage
#            $cmpImage add image -image [image create pixmap -data $image($picName)]
#            $cmpImage add space -width 8
#            $cmpImage add text   -text $menuText

#            set label "-image $cmpImage"
#        }

        eval $menuName add command $label $args

        return
    }

    # mainFrame::CreateMenu --
    #
    #        Exit from TCL Developer Studio
    #
    #        Arguments: (no arguments)
    #
    #        Results: (no results)
    proc CreateMenu {} {
        global image

        variable MainFrame
        variable MenuFrame
        variable RecentFilesMenu
        variable RecentProjectsMenu

        #create frame for menu
        set MenuFrame $MainFrame.menu
        menu $MenuFrame -tearoff 0

        $MenuFrame add cascade -label "File" -underline 0 -menu $MenuFrame.file
        help::add $MenuFrame,0 "File Operations"
        
        menu $MenuFrame.file -tearoff 0

        CreateMenuItem $MenuFrame.file "New..." "vc,newText" -underline 0  -command "mainFrame::onFileNew" -accelerator "Ctrl-N"
        help::add "$MenuFrame.file,0" "Create a new file or project"

        CreateMenuItem $MenuFrame.file "Open" "vc,open" -underline 0 -command "mainFrame::onFileOpen" -accelerator "Ctrl-O"
        help::add "$MenuFrame.file,1" "Open an existing file"

        CreateMenuItem $MenuFrame.file "Close..." "" -underline 0 -state disabled -command "mainFrame::OnFileClose"
        help::add "$MenuFrame.file,2" "Close the file"
        $MenuFrame.file add separator

        CreateMenuItem $MenuFrame.file "Save" "vc,save" -underline 0 -state disabled -command "mainFrame::onFileSave" -accelerator "Ctrl-S"
        help::add "$MenuFrame.file,4" "Saves the file"

        CreateMenuItem $MenuFrame.file "Save As..." "" -underline 5 -state disabled -command "mainFrame::OnFileSaveAs"
        help::add "$MenuFrame.file,5" "Saves the file with new name"

        CreateMenuItem $MenuFrame.file "Save All" "vc,saveAll" -underline 6 -state disabled -command "mainFrame::OnSaveAll"
        help::add "$MenuFrame.file,6" "Saves all the open files"

        CreateMenuItem $MenuFrame.file "Print" "vc,print" -underline 0 -state disabled -command "mainFrame::onFilePrint" -accelerator "Ctrl-P"
        help::add "$MenuFrame.file,7" "Prints current file"

        $MenuFrame.file add separator

        set RecentFilesMenu $MenuFrame.file.recentFiles
        $MenuFrame.file add cascade -label "Recent Files" -menu $RecentFilesMenu

        set RecentProjectsMenu $MenuFrame.file.recentProjects
        $MenuFrame.file add cascade -label "Recent Projects" -menu $RecentProjectsMenu

        UpdateRecentFilesMenu
        UpdateRecentProjectsMenu

        $MenuFrame.file add separator

        CreateMenuItem $MenuFrame.file "Exit" "" -underline 1 -command "mainFrame::exit"
        help::add "$MenuFrame.file,11" "Quits the application; prompts to save files"

        $MenuFrame add cascade -label "Edit" -underline 0 -menu $MenuFrame.edit
        help::add $MenuFrame,1 "Edit Operations"
        
        menu $MenuFrame.edit -tearoff 0

        CreateMenuItem $MenuFrame.edit "Undo" "undo" -underline 0 -state disabled -command "editorWindows::undo" -accelerator "Ctrl-Z"
        help::add "$MenuFrame.edit,0" "Undoes the last action"

        CreateMenuItem $MenuFrame.edit "Redo" "redo" -underline 0 -state disabled -command "editorWindows::redo" -accelerator "Ctrl-Y"
        help::add "$MenuFrame.edit,1" "Redoes the previously undone action"

        $MenuFrame.edit add separator

        CreateMenuItem $MenuFrame.edit "Cut" "vc,cut" -underline 2 -state disabled -command "mainFrame::OnEditCut" -accelerator "Ctrl-X"
        help::add "$MenuFrame.edit,3" "Cuts the selection and moves it to the Clipboard"

        CreateMenuItem $MenuFrame.edit "Copy" "vc,copy" -underline 0 -state disabled -command "mainFrame::OnEditCopy" -accelerator "Ctrl-C"
        help::add "$MenuFrame.edit,4" "Copies the selection to the Clipboard"

        CreateMenuItem $MenuFrame.edit "Paste" "vc,paste" -underline 0 -state disabled -command "mainFrame::OnEditPaste" -accelerator "Ctrl-V"
        help::add "$MenuFrame.edit,5" "Inserts the Clipboard contents at the insertion point"

        CreateMenuItem $MenuFrame.edit "Delete" "vc,delete" -underline 0 -state disabled -command "mainFrame::OnEditDelete" -accelerator "Del"
        help::add "$MenuFrame.edit,6" "Deletes the selection"

        $MenuFrame.edit add separator

        CreateMenuItem $MenuFrame.edit "Select All" "" -underline 8 -state disabled -command "editorWindows::selectAll" -accelerator "Ctrl-A"
        help::add "$MenuFrame.edit,8" "Selects the entire file"

        $MenuFrame.edit add separator

        CreateMenuItem $MenuFrame.edit "Find" "vc,find" -underline 0 -state disabled -command "mainFrame::OnEditFind" -accelerator "Ctrl-F"
        help::add "$MenuFrame.edit,10" "Finds the specified text"

        CreateMenuItem $MenuFrame.edit "Find Next" "vc,findNext" -underline 0 -state disabled -command "mainFrame::OnEditFindNext" -accelerator "Ctrl-L"
        help::add "$MenuFrame.edit,11" "Finds the next occurience"

        CreateMenuItem $MenuFrame.edit "Find in Files" "vc,findInFiles" -underline 1 -command "mainFrame::OnEditFindInFiles"
        help::add "$MenuFrame.edit,12" "Searches for a string in multiple files"

        CreateMenuItem $MenuFrame.edit "Replace" "" -underline 1 -state disabled -command "mainFrame::OnEditReplace" -accelerator "Ctrl-H"
        help::add "$MenuFrame.edit,13" "Replaces the specified text with different text"

        $MenuFrame.edit add separator

        CreateMenuItem $MenuFrame.edit "Go To..." "" -underline 0 -state disabled -accelerator "Ctrl-G" -command "mainFrame::goTo"
        help::add "$MenuFrame.edit,15" "Moves to a specified location"

        $MenuFrame.edit add separator

        $MenuFrame.edit add cascade -label "Advanced" -underline 0 -menu $MenuFrame.edit.adv -state disabled

        menu $MenuFrame.edit.adv -tearoff 0
        $MenuFrame.edit.adv add command -label "Format Selection" -underline 0 -command editorWindows::formatSelection -accelerator "Ctrl-Shift-F"
        $MenuFrame.edit.adv add command -label "Comment Selection" -underline 0 -command editorWindows::commentSelection -accelerator "Ctrl-Shift-C"
        $MenuFrame.edit.adv add command -label "Tabify Selection" -underline 0 -command editorWindows::tabifySelection
        $MenuFrame.edit.adv add command -label "Untabify Selection" -underline 1 -command editorWindows::untabifySelection
        $MenuFrame.edit.adv add command -label "Make Selection Uppercase" -underline 15 -command editorWindows::uppercaseSelection
        $MenuFrame.edit.adv add command -label "Make Selection Lowercase" -underline 15 -command editorWindows::lowercaseSelection
        $MenuFrame.edit.adv add command -label "Strip Selection" -underline 0 -command editorWindows::stripSelection
        $MenuFrame.edit.adv add command -label "Indent Selection" -underline 0 -command editorWindows::indentSelection -accelerator "Ctrl-Shift-I"
        $MenuFrame.edit.adv add command -label "Unindent Selection" -underline 0 -command editorWindows::unindentSelection -accelerator "Ctrl-Shift-U"

        $MenuFrame add cascade -label "Project" -underline 0 -menu $MenuFrame.project
        help::add $MenuFrame,2 "Project Operations"
        menu $MenuFrame.project -tearoff 0

        CreateMenuItem $MenuFrame.project "New" "vc,newFolder" -underline 0 -command "mainFrame::NewProject"
        help::add "$MenuFrame.project,0" "Creates new project"

        CreateMenuItem $MenuFrame.project "Open" "vc,open" -underline 0 -command "mainFrame::onFileOpen"
        help::add "$MenuFrame.project,1" "Opens existing project"

        CreateMenuItem $MenuFrame.project "Save" "vc,save" -underline 0 -command "mainFrame::OnProjectSave" -state disabled
        help::add "$MenuFrame.project,2" "Closes current project"

        CreateMenuItem $MenuFrame.project "Close" "" -underline 0 -command "mainFrame::OnProjectClose" -state disabled
        help::add "$MenuFrame.project,3" "Close current project"

        $MenuFrame.project add separator

        CreateMenuItem $MenuFrame.project "Run" "vc,exec" -underline 0 -command "mainFrame::OnRun" -state disabled
        help::add "$MenuFrame.project,5" "Executes the program"

        CreateMenuItem $MenuFrame.project "Debug" "vc,run" -underline 0 -command "mainFrame::OnDebug" -state disabled
        help::add "$MenuFrame.project,6" "Starts the debuger"


        $MenuFrame.project add separator
        $MenuFrame.project add command -label "Settings..." -underline 0 -command "project::runDialog; mainFrame::enable Run; mainFrame::enable Debug"
        help::add "$MenuFrame.project,12" "Edits the project settings"

        $MenuFrame add cascade -label "Tools" -underline 0 -menu $MenuFrame.tools
        help::add $MenuFrame,3 "Tools Operations"

        menu $MenuFrame.tools -tearoff 0
        $MenuFrame.tools add command -label "Options" -command "options::runDialog"
        help::add "$MenuFrame.tools,0" "Changes the options"

        $MenuFrame add cascade -label "Help" -underline 0 -menu $MenuFrame.help
        help::add $MenuFrame,4 "Help Operations"
        menu $MenuFrame.help -tearoff 0

        CreateMenuItem $MenuFrame.help "About" "" -underline 0 -command "mainFrame::OnAbout"
        help::add "$MenuFrame.help,0" "About TCL Developer Studio"

        return
    }


    # mainFrame::CreateToolBar --
    #
    #        Exit from TCL Developer Studio
    #
    #        Arguments: (no arguments)
    #
    #        Results: (no results)
    proc CreateToolBar {} {

        global image
        global tclDevData

        variable MainFrame
        variable ToolbarFrame

        #create frame for toolbar
        set ToolbarFrame $MainFrame.toolbar
        frame $ToolbarFrame
        #    -relief raised -borderwidth 1
        pack $ToolbarFrame -side top -anchor w -fill x

        frame $ToolbarFrame.sub1 -relief raised -borderwidth 1
        pack $ToolbarFrame.sub1 -side left -fill y
        button $ToolbarFrame.sub1.filenew -image [image create pixmap -data $image(vc,newText)] -relief flat -padx 0 -pady 0 -command "mainFrame::NewTCLFile"
        help::add $ToolbarFrame.sub1.filenew "New TCL File"
        pack $ToolbarFrame.sub1.filenew -side left

        frame $ToolbarFrame.sub2 -relief raised -borderwidth 1
        pack $ToolbarFrame.sub2 -side left -fill y
        button $ToolbarFrame.sub2.fileopen -image [image create pixmap -data $image(vc,open)] -relief flat -padx 0 -pady 0 -command "mainFrame::onFileOpen"
        help::add $ToolbarFrame.sub2.fileopen "Open"
        button $ToolbarFrame.sub2.filesave -image [image create pixmap -data $image(vc,save)] -relief flat -padx 0 -pady 0 -command "mainFrame::onFileSave" -state disabled
        help::add $ToolbarFrame.sub2.filesave "Save"
        button $ToolbarFrame.sub2.filesaveAll -image [image create pixmap -data $image(vc,saveAll)] -relief flat -padx 0 -pady 0 -command "mainFrame::OnSaveAll" -state disabled
        help::add $ToolbarFrame.sub2.filesaveAll "Save All files"
        button $ToolbarFrame.sub2.filePrint -image [image create pixmap -data $image(vc,print)] -relief flat -padx 0 -pady 0 -command "mainFrame::onFilePrint" -state disabled
        help::add $ToolbarFrame.sub2.filesaveAll "Print current file"
        pack $ToolbarFrame.sub2.fileopen $ToolbarFrame.sub2.filesave $ToolbarFrame.sub2.filesaveAll $ToolbarFrame.sub2.filePrint -side left

        frame $ToolbarFrame.sub3 -relief raised -borderwidth 1
        pack $ToolbarFrame.sub3 -side left -fill y
        button $ToolbarFrame.sub3.editcut -image [image create pixmap -data $image(vc,cut)] -relief flat -padx 0 -pady 0  -command "mainFrame::OnEditCut" -state disabled
        help::add $ToolbarFrame.sub3.editcut "Cut"
        button $ToolbarFrame.sub3.editcopy -image [image create pixmap -data $image(vc,copy)] -relief flat -padx 0 -pady 0 -command "mainFrame::OnEditCopy" -state disabled
        help::add $ToolbarFrame.sub3.editcopy "Copy"
        button $ToolbarFrame.sub3.editpaste -image [image create pixmap -data $image(vc,paste)] -relief flat -padx 0 -pady 0 -command "mainFrame::OnEditPaste" -state disabled
        help::add $ToolbarFrame.sub3.editpaste "Paste"
        pack $ToolbarFrame.sub3.editcut $ToolbarFrame.sub3.editcopy $ToolbarFrame.sub3.editpaste -side left

        frame $ToolbarFrame.sub4 -relief raised -borderwidth 1
        pack $ToolbarFrame.sub4 -side left -fill y
        button $ToolbarFrame.sub4.undo -image [image create pixmap -data $image(undo)] -relief flat -padx 0 -pady 0 -state disabled -command "editorWindows::undo"
        help::add $ToolbarFrame.sub4.undo "Undo"
        button $ToolbarFrame.sub4.redo -image [image create pixmap -data $image(redo)] -relief flat -padx 0 -pady 0 -state disabled -command "editorWindows::redo"
        help::add $ToolbarFrame.sub4.redo "Redo"
        pack $ToolbarFrame.sub4.undo $ToolbarFrame.sub4.redo -side left

        frame $ToolbarFrame.sub5 -relief raised -borderwidth 1
        pack $ToolbarFrame.sub5 -side left -fill y
        button $ToolbarFrame.sub5.find -image [image create pixmap -data $image(vc,findInFiles)] -relief flat -padx 0 -pady 0 -command mainFrame::OnFindButton
        help::add $ToolbarFrame.sub5.find "Find in Files"
        tixComboBox $ToolbarFrame.sub5.findStr -editable 1
        pack $ToolbarFrame.sub5.find $ToolbarFrame.sub5.findStr -side left
        #create binding for updating listbox
        trace variable tclDevData(find) w mainFrame::UpdateFindStrAndItem
        bind [$ToolbarFrame.sub5.findStr subwidget entry] <KeyRelease-Return> "mainFrame::OnFindEntry; break"

        frame $ToolbarFrame.sub6 -relief raised -borderwidth 1
        pack $ToolbarFrame.sub6 -side left -fill y
        button $ToolbarFrame.sub6.debug -image [image create pixmap -data $image(vc,run)] -relief flat -padx 0 -pady 0 -state disabled -command "mainFrame::OnDebug"
        help::add $ToolbarFrame.sub6.debug "Start debugger"
        button $ToolbarFrame.sub6.exec -image [image create pixmap -data $image(vc,exec)] -relief flat -padx 0 -pady 0 -state disabled -command "mainFrame::OnRun"
        help::add $ToolbarFrame.sub6.exec "Run Program"
        pack $ToolbarFrame.sub6.debug $ToolbarFrame.sub6.exec -side left

        frame $ToolbarFrame.sub7 -relief raised -borderwidth 1
        pack $ToolbarFrame.sub7 -side left -fill both -expand 1
		menubutton $ToolbarFrame.sub7.windows -image [image create pixmap -data $image(vc,windows)] -relief flat -padx 0 -pady 0 -menu $ToolbarFrame.sub7.windows.m
        menu $ToolbarFrame.sub7.windows.m -tearoff 0
		pack $ToolbarFrame.sub7.windows -side left
		FillWindowsMenu
		trace variable tclDevData(files) w mainFrame::FillWindowsMenu
        help::add $ToolbarFrame.sub7.windows "Windows List"
        
        frame $ToolbarFrame.sub8 -relief raised -borderwidth 1
        pack $ToolbarFrame.sub8 -side right -fill y
        button $ToolbarFrame.sub8.close -image [image create pixmap -data $image(vc,delete)] -relief flat -padx 0 -pady 0 -state disabled -command "mainFrame::OnFileClose"
        help::add $ToolbarFrame.sub7.close "Close file"
        pack $ToolbarFrame.sub8.close

        return
    }

    proc UpdateFindStrAndItem {name1 name2 op} {
        global tclDevData
        variable ToolbarFrame

        set lb [$ToolbarFrame.sub5.findStr subwidget listbox]

        # delete all entries in listbox
        $lb delete 0 end

        foreach l $tclDevData(find) {
            $lb insert end $l
        }

        if {$tclDevData(find) != {}} {
            $ToolbarFrame.sub5.findStr configure -value [lindex $tclDevData(find) 0]
            enable EditFindNext
        }
        return
    }

    # mainFrame::CreateStatusBar --
    #
    #        Sreate Status Bar at the bottom of the Main Frame.
    #        Status bar used for displaying status string
    #        and other usefull information
    #
    #        Arguments: (no arguments)
    #
    #        Results: (no results)

    proc CreateStatusBar {} {
        global tclDevData

        variable MainFrame
        variable StatusFrame

        set StatusFrame [statusText::create $MainFrame]

        #create cursor position labels
        set tclDevData(cursor,line) 0
        set tclDevData(cursor,pos) 0

        label $StatusFrame.line -textvariable tclDevData(cursor,line) -width 6
        label $StatusFrame.d -text ":"
        label $StatusFrame.pos -textvariable tclDevData(cursor,pos) -width 6

        pack $StatusFrame.pos $StatusFrame.d $StatusFrame.line -side right

        bind $StatusFrame.line <Button> mainFrame::OnLineClick

        # create time field
        label $StatusFrame.time -width 5
        pack $StatusFrame.time -padx 3 -side right

        after 1000 mainFrame::UpdateTimeField

        return
    }

    proc UpdateTimeField {} {
        variable StatusFrame

        $StatusFrame.time configure -text [clock format [clock seconds] -format "%H:%M"]

        after 1000 mainFrame::UpdateTimeField

        return
    }

    proc OnLineClick {} {
        variable StatusFrame
        global tclDevData

        if {$tclDevData(curFile) != ""} {
            mainFrame::goTo [$StatusFrame.line cget -text]
        }

        return
    }

    # mainFrame::OnFileOpen --
    #
    #               Reaction on FileOpen Command.
    #        Show Open File dialog
    #        Arguments: (no arguments)
    #
    #        Results: (no results)
    proc onFileOpen {} {
        global tclDevData

        # use native open dialog
        set types {
            {{TCL Scripts} {.tcl}}
            {{All files}    *    }
        }

        set fileName [tk_getOpenFile -filetypes $types]

        if {$fileName != ""} {
            openFile $fileName
        }
    }

    proc openFile {fileName} {
        set ext [file extension $fileName]
        switch $ext {
            .tcl {
                OpenTCLFile $fileName
            }

            default {
                openTextFile $fileName
            }
        }

        return
    }

    # This procedure simple open text file
    proc openTextFile { fileName {isProject 0}} {
        global tclDevData

        if {[lsearch $tclDevData(files) $fileName] == -1} {
            # load new file
            LoadFile $fileName $isProject

            projectWindow::addFile $fileName $isProject
            editorWindows::openFile $fileName

            lappend tclDevData(files) $fileName

            if {!$isProject} {
                AddToRecentFiles $fileName
            }
        } else {
            # select already loaded file
            projectWindow::selectFile $fileName
            editorWindows::selectFile $fileName
        }

        return
    }

    # this procedure determine, is it project file and open file as project, or as tcl file
    proc OpenTCLFile { fileName } {
        set isProject 0

        if {[IsProjectFile $fileName] &&
            [tk_messageBox -icon question \
            -type yesno \
            -title "TCL DevStudio" \
            -message "Load $fileName as TCLDevStudio Project File?"] == "yes"} {
            if {![OpenProject $fileName]} {
                return 0
            }

            set isProject 1
        }

        # not implemented yet.
        openTextFile $fileName $isProject

        editorWindows::colorizeFile $fileName
        return
    }

    # Open TCL DevStudio Project
    proc OpenProject { fileName } {
        global tclDevData

        if {$tclDevData(projectFile) != ""} {
            # close Project
            if {![closeFile $tclDevData(projectFile)]} {
                return 0
            }
        }

        project::loadFile $fileName

        enable Run
        enable Debug
        AddToRecentProjects $fileName
        set tclDevData(projectFile) $fileName

        enable ProjectSave
        enable ProjectClose

        return 1
    }


    # mainFrame::LoadFile --
    #
    #       Load File
    #       Arguments: fileName - name of file for loading
    #
    #    Results: 1 - if OK, 0 - if Error
    proc LoadFile {fileName {isProject 0}} {
        global tclDevData

        if {[catch {set fileID [open $fileName "r"]}]} {
            error "Can not load file"
        }

        lappend tclDevData(files,$fileName)

        set tclDevData(files,$fileName,text) [read $fileID]
        if {![info exist tclDevData(files,$fileName,cursor)]} {
            set tclDevData(files,$fileName,cursor) 1.0
        }
        if {![info exist tclDevData(files,$fileName,anchor)]} {
            set tclDevData(files,$fileName,anchor) ""
        }
        if {![info exist tclDevData(files,$fileName,selFirst)]} {
            set tclDevData(files,$fileName,selFirst) ""
        }
        if {![info exist tclDevData(files,$fileName,selLast)]} {
            set tclDevData(files,$fileName,selLast) ""
        }
        set tclDevData(files,$fileName,isUntitled) 0
        set tclDevData(files,$fileName,isChanged) 0
        set tclDevData(files,$fileName,marks) {}
        set tclDevData(files,$fileName,keywords) {}
        set tclDevData(files,$fileName,comments) {}
        set tclDevData(files,$fileName,isProject) $isProject

        close $fileID

        # convert all tabs to spaces
        #    if {$tclDevData(options,changeTabs)} {
            #        set index [string first "\t" $tclDevData(files,$fileName,text)]
            #
            #        set spaces ""
            #        for {set i 0} {$i < $tclDevData(options,tabSize)} {incr i} {
                #            append spaces " "
                #        }
            #
            #        while {$index != -1} {
                #            set text [string range $tclDevData(files,$fileName,text) 0 [expr {$index - 1}]]
                #            append text $spaces
                #            append text [string range $tclDevData(files,$fileName,text) [expr {$index + 1}] end]
                #
                #            set tclDevData(files,$fileName,text) $text
                #
                #            set index [string first "\t" $tclDevData(files,$fileName,text)]
                #        }
            #    }

        # delete all TCLDEV Comments
        if {$isProject} {
            set text [split $tclDevData(files,$fileName,text) "\n"]
            set lineNum 0
            set lines [llength $text]

            set line ""

            #find start of project settings
            while {$lineNum < $lines} {
                set line [lindex $text $lineNum]

                if {$line == "##TCLDevStudioProject"} {
                    break
                }
                incr lineNum
            }

            if {$line != "##TCLDevStudioProject"} {
                return 1
            }

            #delete current line
            set text [lreplace $text $lineNum $lineNum]
            incr lines -1

            #copy all lines to temp file
            while {$lineNum < $lines} {
                set line [lindex $text $lineNum]

                #is it end of properties?
                if {$line == "##TCLDevStudioProject_END"} {
                    break
                }
                #is it comment?
                if {[string index $line 0] != "#"} {
                    #this is not comment
                    break
                }

                #delete current line
                set text [lreplace $text $lineNum $lineNum]
                incr lines -1
            }

            #process last line
            if {$line == "##TCLDevStudioProject_END"} {
                #delete current line
                set text [lreplace $text $lineNum $lineNum]
                incr lines -1

                set tclDevData(files,$fileName,text) [join $text "\n"]
                return 1
            } else {
                return 1
            }

        }

        return 1
    }

    # If file untitled, ask name and then save it on disk
    proc SaveFile { fileName } {
        global tclDevData

        if {$tclDevData(files,$fileName,isUntitled)} {
            return [saveAs $fileName]
        } else {
            return [storeFile $fileName]
        }
    }


    # Save file on disk
    proc storeFile { fileName } {
        global tclDevData

        if {[catch {set fileID [open $fileName "w"]}]} {
            return 0
        }

        fconfigure $fileID -translation lf -buffering full -buffersize 100000

        puts -nonewline $fileID $tclDevData(files,$fileName,text)

        # is it project?
        if {$tclDevData(files,$fileName,isProject)} {
            #store project settings in the end  of file
            puts $fileID ""
            project::save $fileID
            AddToRecentProjects $fileName
        } else  {
            AddToRecentFiles $fileName
        }

        close $fileID

        set tclDevData(files,$fileName,isUntitled) 0
        set tclDevData(files,$fileName,isChanged) 0

        return 1
    }

    # Ask new name for file and save it
    proc saveAs {fileName} {
        global tclDevData

        set types {
            {{TCL Scripts} {.tcl}}
            {{All files}    *    }
        }

        set newName [tk_getSaveFile -filetypes $types \
        -title "Save $fileName as..."]

        if {$newName == ""} {
            return 0
        } elseif {[lsearch $tclDevData(files) $newName] != -1} {
            tk_messageBox -title "TCL Developer Studio" -message "File $newName already edited." -icon error
            return 0
        } else {
            RenameFile $fileName $newName

            return [storeFile $newName]
        }
    }

    # close file
    proc closeFile { fileName } {
        global tclDevData

        if {$fileName == $tclDevData(curFile)} {
            editorWindows::applyChanges
        }

        #always save project
        if {$tclDevData(files,$fileName,isProject)} {
            set reply [tk_messageBox -icon question \
            -type yesnocancel \
            -title "TCL DevStudio" \
            -message "Save Project $fileName?"]

            # process reply
            switch $reply {
                "yes" {
                    # save File
                    if {[SaveFile $fileName] == 0} {
                        return 0
                    }
                }

                "no" {
                    # do not save file
                }

                "cancel" {
                    # do not close file
                    return 0
                }
            }
        } else {
            # is file was changed?
            if {[editorWindows::fileChanged $fileName] ||
                $tclDevData(files,$fileName,isUntitled)} {
                # ask for saving
                set reply [tk_messageBox -icon question \
                -type yesnocancel \
                -title "TCL DevStudio" \
                -message "File $fileName was changed. Save it?"]

                # process reply
                switch $reply {
                    "yes" {
                        # save File
                        if {[SaveFile $fileName] == 0} {
                            return 0
                        }
                    }

                    "no" {
                        # do not save file
                    }

                    "cancel" {
                        # do not close file
                        return 0
                    }
                }
            }
        }

        # make closing

        # search file in list
        set index [lsearch $tclDevData(files) $fileName]
        # check result of searching
        if {$index == -1} {
            error "Internal Error"
        }

        # delete file from list
        set tclDevData(files) [lreplace $tclDevData(files) $index $index]

        # select another file
        if {$tclDevData(files) == {}} {
            set newFile ""
        } else {
            set newFile [lindex $tclDevData(files) 0]
        }

        projectWindow::deleteFile $fileName $newFile
        editorWindows::deleteFile $fileName $newFile

        if {$tclDevData(files,$fileName,isProject)} {
            set tclDevData(projectFile) ""
            project::resetProject

            disable ProjectSave
            disable ProjectClose
        }

        # delete information, associated with this file.
        unset tclDevData(files,$fileName,text)
        unset tclDevData(files,$fileName,isChanged)
        unset tclDevData(files,$fileName,isUntitled)
        unset tclDevData(files,$fileName,marks)
        unset tclDevData(files,$fileName,comments)
        unset tclDevData(files,$fileName,keywords)
        unset tclDevData(files,$fileName,isProject)

        foreach name [array names tclDevData files,$fileName,marks,*] {
            unset tclDevData($name)
        }

        return 1
    }

    # Reaction on Select "Save" menu
    proc onFileSave {} {
        global tclDevData

        if {$tclDevData(curFile) != ""} {
            editorWindows::applyChanges

            SaveFile $tclDevData(curFile)
        }

        return
    }

    # Reaction on select "Save As..." menu
    proc OnSaveAll {} {
        global tclDevData
        set result 1

        # apply changes in current file
        if {$tclDevData(curFile) != ""} {
            editorWindows::applyChanges
        }

        # for all files
        foreach fileName $tclDevData(files) {
            # save file
            if {[editorWindows::fileChanged $fileName] ||
                $tclDevData(files,$fileName,isUntitled)} {
                    if {![SaveFile $fileName]} {
                        set result 0
                    }
            }
        }
        
        return $result
    }

    # Reaction on selecting "Close"
    proc OnFileClose {} {
        global tclDevData

        if {$tclDevData(curFile) != ""} {
            closeFile $tclDevData(curFile)
        }
    }

    # close all files
    proc CloseAll {} {
        global tclDevData

        # apply changes in current file
        if {$tclDevData(curFile) != ""} {
            editorWindows::applyChanges
        }

        #Save Project
        if {$tclDevData(projectFile) != ""} {
            set reply [tk_messageBox -icon question \
            -type yesnocancel \
            -title "TCL DevStudio" \
            -message "Save Project $tclDevData(projectFile)?"]

            # process reply
            switch $reply {
                "yes" {
                    # save File
                    if {[SaveFile $tclDevData(projectFile)] == 0} {
                        return 0
                    }
                }

                "no" {
                    # do not save file
                }

                "cancel" {
                    # do not close file
                    return 0
                }
            }
        }

        # Save All files
        # apply changes in current file
        if {$tclDevData(curFile) != ""} {
            editorWindows::applyChanges
        }

        # for all files
        foreach fileName $tclDevData(files) {
            # is file was changed?
            if {!$tclDevData(files,$fileName,isProject) &&
                ([editorWindows::fileChanged $fileName] ||
                $tclDevData(files,$fileName,isUntitled))} {
                # ask for saving
                set reply [tk_messageBox -icon question \
                -type yesnocancel \
                -title "TCL DevStudio" \
                -message "File $fileName was changed. Save it?"]

                # process reply
                switch $reply {
                    "yes" {
                        # save File
                        if {[SaveFile $fileName] == 0} {
                            return 0
                        }
                    }

                    "no" {
                        # do not save file
                    }

                    "cancel" {
                        # do not close file
                        return 0
                    }
                }
            }
        }

        return 1
    }

    # Reaction on selecting "New" menu
    proc onFileNew {} {
        set result [newDialog::run]

        if {$result != {}} {
            array set newDialogData $result

            switch $newDialogData(what) {
                "file" {
                    NewFile
                }
                "tcl" {
                    NewTCLFile
                }
                "project" {
                    NewProject
                }
            }
        }

        return
    }

    proc NewFile {} {
        global tclDevData

        #generate unique unknown
        for {set i 1} {1} {incr i} {
            #create file name
            set name "untitled"
            append name $i

            if {[lsearch $tclDevData(files) $name] == -1} {
                break
            }
        }

        #create data for this file
        lappend tclDevData(files) $name
        set tclDevData(files,$name,text) ""
        set tclDevData(files,$name,cursor) 0.0
        set tclDevData(files,$name,selFirst) ""
        set tclDevData(files,$name,selLast) ""
        set tclDevData(files,$name,anchor) ""
        set tclDevData(files,$name,isUntitled) 1
        set tclDevData(files,$name,isChanged) 0
        set tclDevData(files,$name,marks) {}
        set tclDevData(files,$name,comments) {}
        set tclDevData(files,$name,keywords) {}
        set tclDevData(files,$name,isProject) 0

        projectWindow::addFile $name
        editorWindows::openFile $name
    }

    proc NewTCLFile {} {
        global tclDevData

        #generate unique unknown
        for {set i 1} {1} {incr i} {
            #create file name
            set name "untitled"
            append name $i
            append name ".tcl"

            if {[lsearch $tclDevData(files) $name] == -1} {
                break
            }
        }

        #create data for this file
        lappend tclDevData(files) $name
        set tclDevData(files,$name,text) [project::getFileComment $name]
        set tclDevData(files,$name,cursor) end
        set tclDevData(files,$name,selFirst) ""
        set tclDevData(files,$name,selLast) ""
        set tclDevData(files,$name,anchor) ""
        set tclDevData(files,$name,isUntitled) 1
        set tclDevData(files,$name,isChanged) 0
        set tclDevData(files,$name,marks) {}
        set tclDevData(files,$name,comments) {}
        set tclDevData(files,$name,keywords) {}
        set tclDevData(files,$name,isProject) 0

        projectWindow::addFile $name
        editorWindows::openFile $name
    }

    proc NewProject {} {
        global tclDevData

        # close prev project
        if {$tclDevData(projectFile) != ""} {
            if {![closeFile $tclDevData(projectFile)]} {
                return
            }
        }

        #generate unique unknown
        for {set i 1} {1} {incr i} {
            #create file name
            set name "untitled"
            append name $i
            append name ".tcl"

            if {[lsearch $tclDevData(files) $name] == -1} {
                break
            }
        }

        #create data for this file
        lappend tclDevData(files) $name
        set tclDevData(files,$name,text) ""
        set tclDevData(files,$name,cursor) 0.0
        set tclDevData(files,$name,selFirst) ""
        set tclDevData(files,$name,selLast) ""
        set tclDevData(files,$name,anchor) ""
        set tclDevData(files,$name,isUntitled) 1
        set tclDevData(files,$name,isChanged) 0
        set tclDevData(files,$name,marks) {}
        set tclDevData(files,$name,comments) {}
        set tclDevData(files,$name,keywords) {}
        set tclDevData(files,$name,isProject) 1

        set tclDevData(projectFile) $name

        projectWindow::addFile $name 1
        editorWindows::openFile $name

        enable ProjectSave
        enable ProjectClose

        return
    }

    #save current project
    proc OnProjectSave {} {
        global tclDevData

        if {$tclDevData(projectFile) == ""} {
            error "Internal Error"
        }

        SaveFile $tclDevData(projectFile)

        return
    }

    #close current project
    proc OnProjectClose {} {
        global tclDevData

        if {$tclDevData(projectFile) == ""} {
            error "Internal Error"
        }

        closeFile $tclDevData(projectFile)

        return
    }

    # Rename file. Used in SaveAs command
    proc RenameFile {oldName newName} {
        global tclDevData

        set index [lsearch $tclDevData(files) $oldName]

        if {$index == -1} {
            error "Internal error"
        }

        set tclDevData(files) [lreplace $tclDevData(files) $index $index $newName]

        # rename all variables
        foreach name [array names tclDevData files,$oldName,*] {
            #change second field in variable
            set newArrayName [join [lreplace [split $name ","] 1 1 $newName] ","]

            set tclDevData($newArrayName) $tclDevData($name)
            unset tclDevData($name)
        }

        #rename curFile
        if {$tclDevData(curFile) == $oldName} {
            set tclDevData(curFile) $newName
        }

        projectWindow::renameFile $oldName $newName $tclDevData(files,$newName,isProject)

        if {$tclDevData(projectFile) == $oldName} {
            set tclDevData(projectFile) $newName
        }
    }

    # Reaction on selecting "Save As" menu
    proc OnFileSaveAs {} {
        global tclDevData

        editorWindows::applyChanges

        if {$tclDevData(curFile) != ""} {
            saveAs $tclDevData(curFile)
        }
    }

    # Reaction on selecting "Run" menu
    proc OnRun {} {
        global tcl_platform
        variable IsRunned

        # save all files
        if {![OnSaveAll]} {
            return
        }

        if {[project::getDir] != ""} {
            cd [project::getDir]
        }

        if {[project::getRunCmd] != ""} {
            if {$tcl_platform(platform) == "windows"} {
                eval exec [project::getRunCmd] &
            } else  {
                disable Run
                disable Debug

                set IsRunned 1

                # Execute
                set cmd [project::getRunCmd]

                # clear output window
                outputWindow::clearDebug

                set outputStream [open "| $cmd |& cat" r+]
                fconfigure $outputStream -buffering none -blocking 0
                fileevent $outputStream readable "outputWindow::onDebugOutput $outputStream"
            }
        }

        return
    }

    # Reaction on selecting "Debug" menu
    proc OnDebug {} {
        variable IsDebugered

        # save all files
        if {![OnSaveAll]} {
            return
        }

        if {[project::getDir] != ""} {
            cd [project::getDir]
        }

        if {[project::getDebugCmd] != ""} {

            disable Run
            disable Debug

            set IsDebugered 1

            # Debug
            set cmd [project::getDebugCmd]
            eval "exec $cmd"
        }

        return
    }

    proc endRun {} {
        variable IsDebugered
        variable IsRunned

        set IsDebugered 0
        set IsRunned 0

        enable Run
        enable Debug

        return
    }

    # Determine, is it file - TCLDevStudio project file
    proc IsProjectFile { fileName } {
        if {[catch {set fileID [open $fileName "r"]}]} {
            error "Can not load project $fileName"
        }

        set line ""
        while {[gets $fileID line] != -1} {
            if { $line == "##TCLDevStudioProject" } {
                close $fileID
                return 1
            }
        }

        close $fileID
        return 0
    }

    # Reaction on edit-cut
    proc OnEditCut {} {
        editorWindows::cut
    }

    # Reaction on edit-copy
    proc OnEditCopy {} {
        editorWindows::copy
    }

    # Reaction on edit-paste
    proc OnEditPaste {} {
        editorWindows::paste
    }

    # Reaction on edit-delete
    proc  OnEditDelete {} {
        editorWindows::delete
    }

    proc enable {what} {
        variable MenuFrame
        variable ToolbarFrame
        variable IsRunned
        variable IsDebugered
        global tclDevData

        switch $what {
            "Run" {
                if {[project::getRunCmd] != "" &&
                    $IsRunned != 1 && $IsDebugered != 1} {
                    $MenuFrame.project entryconfigure 5 -state normal
                    $ToolbarFrame.sub6.exec configure -state normal
                }
            }

            "Debug" {
                if {[project::getDebugCmd] != "" &&
                    $IsRunned != 1 && $IsDebugered != 1} {
                    $MenuFrame.project entryconfigure 6 -state normal
                    $ToolbarFrame.sub6.debug configure -state normal
                }
            }

            "ProjectSave" {
                $MenuFrame.project entryconfigure 2 -state normal
            }

            "ProjectClose" {
                $MenuFrame.project entryconfigure 3 -state normal
            }

            "EditUndo" {
                $MenuFrame.edit entryconfigure 0 -state normal
                $ToolbarFrame.sub4.undo configure -state normal
            }
            "EditRedo" {
                $MenuFrame.edit entryconfigure 1 -state normal
                $ToolbarFrame.sub4.redo configure -state normal
            }

            "EditCut" {
                $MenuFrame.edit entryconfigure 3 -state normal
                $ToolbarFrame.sub3.editcut configure -state normal
            }
            "EditCopy" {
                $MenuFrame.edit entryconfigure 4 -state normal
                $ToolbarFrame.sub3.editcopy configure -state normal
            }
            "EditPaste" {
                $MenuFrame.edit entryconfigure 5 -state normal
                $ToolbarFrame.sub3.editpaste configure -state normal
            }
            "EditDelete" {
                $MenuFrame.edit entryconfigure 6 -state normal
            }
            "EditFind" {
                $MenuFrame.edit entryconfigure 10 -state normal
            }
            "EditFindNext" {
                if {$tclDevData(find) != {}} {
                    $MenuFrame.edit entryconfigure 11 -state normal
                }
            }
            "EditReplace" {
                $MenuFrame.edit entryconfigure 13 -state normal
            }
            "EditGoTo" {
                $MenuFrame.edit entryconfigure 15 -state normal
            }
            "EditAdvanced" {
                $MenuFrame.edit entryconfigure 17 -state normal
            }
            "SelectAll" {
                $MenuFrame.edit entryconfigure 8 -state normal
            }

            "FileClose" {
                $MenuFrame.file entryconfigure 2 -state normal
                $ToolbarFrame.sub8.close configure -state normal
            }
            "FileSave" {
                $MenuFrame.file entryconfigure 4 -state normal
                $ToolbarFrame.sub2.filesave configure -state normal
            }
            "FileSaveAs" {
                $MenuFrame.file entryconfigure 5 -state normal
            }
            "FileSaveAll" {
                $MenuFrame.file entryconfigure 6 -state normal
                $ToolbarFrame.sub2.filesaveAll configure -state normal
            }
            "FilePrint" {
                $MenuFrame.file entryconfigure 7 -state normal
                $ToolbarFrame.sub2.filePrint configure -state normal
            }

        }

        return
    }

    proc disable {what} {
        variable MenuFrame
        variable ToolbarFrame

        switch $what {
            "Run" {
                $MenuFrame.project entryconfigure 5 -state disabled
                $ToolbarFrame.sub6.exec configure -state disabled
            }

            "Debug" {
                $MenuFrame.project entryconfigure 1 -state disabled
                $ToolbarFrame.sub6.debug configure -state disabled
            }

            "ProjectSave" {
                $MenuFrame.project entryconfigure 2 -state disabled
            }

            "ProjectClose" {
                $MenuFrame.project entryconfigure 3 -state disabled
            }

            "EditUndo" {
                $MenuFrame.edit entryconfigure 6 -state disabled
                $ToolbarFrame.sub4.undo configure -state disabled
            }
            "EditRedo" {
                $MenuFrame.edit entryconfigure 1 -state disabled
                $ToolbarFrame.sub4.redo configure -state disabled
            }

            "EditCut" {
                $MenuFrame.edit entryconfigure 3 -state disabled
                $ToolbarFrame.sub3.editcut configure -state disabled
            }
            "EditCopy" {
                $MenuFrame.edit entryconfigure 4 -state disabled
                $ToolbarFrame.sub3.editcopy configure -state disabled
            }
            "EditPaste" {
                $MenuFrame.edit entryconfigure 5 -state disabled
                $ToolbarFrame.sub3.editpaste configure -state disabled
            }
            "EditDelete" {
                $MenuFrame.edit entryconfigure 6 -state disabled
            }
            "EditFind" {
                $MenuFrame.edit entryconfigure 10 -state disabled
            }
            "EditFindNext" {
                $MenuFrame.edit entryconfigure 11 -state disabled
            }
            "EditReplace" {
                $MenuFrame.edit entryconfigure 13 -state disabled
            }
            "EditGoTo" {
                $MenuFrame.edit entryconfigure 15 -state disabled
            }
            "EditAdvanced" {
                $MenuFrame.edit entryconfigure 17 -state disabled
            }
            "SelectAll" {
                $MenuFrame.edit entryconfigure 8 -state disabled
            }

            "FileClose" {
                $MenuFrame.file entryconfigure 2 -state disabled
                $ToolbarFrame.sub8.close configure -state disabled
            }
            "FileSave" {
                $MenuFrame.file entryconfigure 4 -state disabled
                $ToolbarFrame.sub2.filesave configure -state disabled
            }
            "FileSaveAs" {
                $MenuFrame.file entryconfigure 5 -state disabled
            }
            "FileSaveAll" {
                $MenuFrame.file entryconfigure 6 -state disabled
                $ToolbarFrame.sub2.filesaveAll configure -state disabled
            }
            "FilePrint" {
                $MenuFrame.file entryconfigure 7 -state disabled
                $ToolbarFrame.sub2.filePrint configure -state disabled
            }
        }

        return

    }

    proc OnFindEntry {} {
        global tclDevData
        variable ToolbarFrame

        set findStr [[$ToolbarFrame.sub5.findStr subwidget entry] get]

        if {$findStr != ""} {
            # append into the find list
            if {[llength $tclDevData(find)] == 0 ||
                $findStr != [lindex $tclDevData(find) 0]} {
                set tclDevData(find) [linsert $tclDevData(find) 0 $findStr]
            }
            #find
            set result [editorWindows::find $findStr]
        }
        return
    }

    proc OnFindButton {} {
        variable ToolbarFrame

        set text [[$ToolbarFrame.sub5.findStr subwidget entry] get]

        findInFiles $text

        return
    }

    proc onCreateText {} {
        enable EditUndo
        enable EditRedo
        enable EditCut
        enable EditCopy
        enable EditPaste
        enable EditDelete
        enable EditFind
        enable EditFindNext
        enable EditReplace
        enable EditGoTo
        enable SelectAll
        enable FileClose
        enable FileSave
        enable FileSaveAs
        enable FileSaveAll
        enable FilePrint
        enable EditAdvanced
        return
    }

    proc onDestroyText {} {
        disable EditUndo
        disable EditRedo
        disable EditCut
        disable EditCopy
        disable EditPaste
        disable EditDelete
        disable EditFind
        disable EditFindNext
        disable EditReplace
        disable EditGoTo
        disable SelectAll
        disable FileClose
        disable FileSave
        disable FileSaveAs
        disable FileSaveAll
        disable FilePrint
        disable EditAdvanced

        return
    }

    # reaction of selection "Find" menu
    proc OnEditFind {} {
        find
        return
    }

    proc OnEditFindInFiles {} {
        findInFiles
        return
    }

    proc OnEditFindNext {} {
        editorWindows::findNext
        return
    }

    proc find {{str ""}}  {
        global tclDevData

        set options ""

        set result [findDialog $str]

        set findStr ""

        if {[lindex $result 0] != ""} {
            set findStr [lindex $result 0]
            set options [lindex $result 1]
        }

        if {$findStr != ""} {
            # append into the find list
            if {[llength $tclDevData(find)] == 0 ||
                $findStr != [lindex $tclDevData(find) 0]} {
                set tclDevData(find) [linsert $tclDevData(find) 0 $findStr]
            }
            set tclDevData(find,lastOptions) $options

            #find
            editorWindows::find $findStr $options
        }

        return
    }

    proc findInFiles {{findStr ""}} {
        global tclDevData

        set result [findInFilesDialog $findStr]

        if {$result != {}} {
            set findStr [lindex $result 0]
            set path [lindex $result 1]
            set fileMasks [lindex $result 2]
            set options [lindex $result 3]

            if {$findStr != ""} {
                # append into the find list
                if {[llength $tclDevData(find)] == 0 ||
                    $findStr != [lindex $tclDevData(find) 0]} {
                    set tclDevData(find) [linsert $tclDevData(find) 0 $findStr]
                }

                # call grep
                set tclDevData(findInFiles) [eval "grep \"$findStr\" \"$path\" $fileMasks $options"]
                statusText::setText "[llength $tclDevData(findInFiles)] occurrence(s) have been found."

                #clear old result
                outputWindow::clearFind

                # apply result
                foreach entry $tclDevData(findInFiles) {
                    outputWindow::addFind [lindex $entry 0] [lindex $entry 1] [lindex $entry 2] [lindex $entry 3]
                }

            }
        }

        return
    }

    # reaction on Replace menu
    proc OnEditReplace {} {
        replace
        return
    }

    proc replace {{findStr ""}} {
        global tclDevData

        set result [replaceDialog $findStr]

        if {$result != {}} {
            set op [lindex $result 0]
            set findStr [lindex $result 1]
            set replaceStr [lindex $result 2]
            set in [lindex $result 3]
            set options [lindex $result 4]

            # store find and replace strings
            if {$tclDevData(find) != {} ||
                $findStr != [lindex $tclDevData(find) 0]} {
                set tclDevData(find) [linsert $tclDevData(find) 0 $findStr]
            }
            if {$tclDevData(replace) != {} ||
                $replaceStr != [lindex $tclDevData(replace) 0]} {
                set tclDevData(replace) [linsert $tclDevData(replace) 0 $replaceStr]
            }

            switch $op {
                "find" {
                    set tclDevData(find,lastOptions) $options

                    #find
                    editorWindows::find $findStr $options
                }

                "replace" {
                    editorWindows::replace $findStr $replaceStr $in $options
                }

                "replaceAll" {
                    editorWindows::replaceAll $findStr $replaceStr $in $options
                }
            }
        }

        return
    }

    proc goTo {{lineNum ""}} {
        set result [goToDialog $lineNum]

        if {$result != {}} {
            set what [lindex $result 0]
            set value [lindex $result 1]

            switch $what {
                "line" {
                    editorWindows::setCursor $value 0
                }

                "procedure" {
                    editorWindows::gotoProc $value
                }
            }
        }

        return
    }

    proc getGeometry {} {
        variable MainFrame

        return [wm geometry $MainFrame]
    }

    proc getProjectWinSize {} {
        variable PaneWindow2

        return [$PaneWindow2 panecget project -size]
    }

    proc getOutputWinSize {} {
        variable PaneWindow1

        return [$PaneWindow1 panecget outWin -size]
    }

    proc AddToRecentFiles {fileName} {
        global tclDevData

        if {![info exist tclDevData(recentFiles)]} {
            lappend tclDevData(recentFiles) $fileName
        }

        #delete prev instance in recent files
        set index [lsearch $tclDevData(recentFiles) $fileName]

        if {$index != -1} {
            set tclDevData(recentFiles) [lreplace $tclDevData(recentFiles) $index $index]
        }

        #insert to begin of list
        set tclDevData(recentFiles) [linsert $tclDevData(recentFiles) 0 $fileName]

        #check size of list
        if {[llength $tclDevData(recentFiles)] > 20} {
            #delete last entry
            set lastIndex [expr {[llength $tclDevData(recentFiles)] - 1}]
            set tclDevData(recentFiles) [lreplace $tclDevData(recentFiles) $lastIndex $lastIndex]
        }

        UpdateRecentFilesMenu

        return
    }

    proc AddToRecentProjects {fileName} {
        global tclDevData

        if {![info exist tclDevData(recentProjects)]} {
            lappend tclDevData(recentProjects) $fileName
        }

        #delete prev instance in recent files
        set index [lsearch $tclDevData(recentProjects) $fileName]

        if {$index != -1} {
            set tclDevData(recentProjects) [lreplace $tclDevData(recentProjects) $index $index]
        }

        #insert to begin of list
        set tclDevData(recentProjects) [linsert $tclDevData(recentProjects) 0 $fileName]

        #check size of list
        if {[llength $tclDevData(recentProjects)] > 20} {
            #delete last entry
            set lastIndex [expr {[llength $tclDevData(recentProjects)] - 1}]
            set tclDevData(recentProjects) [lreplace $tclDevData(recentProjects) $lastIndex $lastIndex]
        }

        UpdateRecentProjectsMenu
        return
    }

    proc UpdateRecentFilesMenu {} {
        variable RecentFilesMenu
        global tclDevData

        catch {destroy $RecentFilesMenu}

        menu $RecentFilesMenu -tearoff 0

        if {[info exist tclDevData(recentFiles)]} {
            foreach fileName $tclDevData(recentFiles) {
                $RecentFilesMenu add command -label $fileName -command "mainFrame::openFile $fileName"
            }
        }

        return
    }

    proc UpdateRecentProjectsMenu {} {
        variable RecentProjectsMenu
        global tclDevData

        catch {destroy $RecentProjectsMenu}

        menu $RecentProjectsMenu -tearoff 0

        if {[info exist tclDevData(recentProjects)]} {
            foreach fileName $tclDevData(recentProjects) {
                $RecentProjectsMenu add command -label $fileName -command "mainFrame::openFile $fileName"
            }
        }

        return
    }

    proc setProjectName {name1 name2 op} {
        variable MainFrame

        if {$MainFrame != ""} {
            if {[project::getName] != ""} {
                wm title $MainFrame "TCL Developer Studio - [project::getName]"
            } else {
                wm title $MainFrame "TCL Developer Studio"
            }
        }

        return
    }

    proc refreshTree {} {
        editorWindows::applyChanges
    }

    proc OnAbout {} {
        global aboutDialog

        set w [toplevel .about -background blue4]
        wm title $w "About"

        titlePane::create $w
        button $w.ok -text "OK" -command "set aboutDialog 1"
        pack $w.ok -side bottom -pady 3

        tkwait visibility $w
        focus $w.ok
        grab $w
        tkwait variable aboutDialog
        grab release $w

        destroy $w

        return
    }
    
    proc FillWindowsMenu {{name1 ""} {name2 ""} {op ""}} {
        global tclDevData
        variable ToolbarFrame
        
        $ToolbarFrame.sub7.windows.m delete 0 end
        
        foreach file $tclDevData(files) {
            $ToolbarFrame.sub7.windows.m add command -label $file -command "editorWindows::selectFile $file"
        }
        
        return
    }
    
    proc selectMenu {index} {
        variable MenuFrame
        
        $MenuFrame invoke $index
        
        return
    }
    
    proc onFilePrint {} {
        global tclDevData
        
        set result [printDialog $tclDevData(curFile)]
        
        #create options
        if {$result != {}} {
            set command $tclDevData(options,printCommand)
            
            if {[lindex $result 0] == "printer"} {
                append command " --printer=[lindex $result 1]"
            } else  {
                append command " --output=[lindex $result 1]"
            }
            
            if {[file extension $tclDevData(curFile)] == "tcl"} {
                append command " --pretty-print=tcl"
            }
            
            if {$tclDevData(options,prettyOutput) == 1} {
                append command " -g"
            }
            
            append command " --stdin=$tclDevData(curFile)"
            
            if {$tclDevData(options,showLineNumbers) == 1} {
                append command " --line-numbers=$tclDevData(options,afterLineNum)"
            }
            
            append command " -1"
            
            # apply changes
            editorWindows::applyChanges
            
            if [catch {eval exec -- $command [list << $tclDevData(files,$tclDevData(curFile),text)]} result] {
                outputWindow::addTextToDebug $result    
           }
        }
        
        return
    }
}

# Go To dialog
proc goToDialog { {lineNum ""}} {
    global goToDialog
    global tclDevData

    set w [toplevel .goToDialog -borderwidth 10]
    wm title $w "Go To:"

    if {[info exist tclDevData(positions,goToDialog)] &&
        [string range $tclDevData(positions,goToDialog) 0 2] != "1x1"} {
        wm geometry $w $tclDevData(positions,goToDialog)
    } else {
        set x [expr {([winfo screenwidth .] - 200)/2}]
        set y [expr {([winfo screenheight .] - 50)/2}]

        wm geometry $w 400x100+$x+$y
    }
    wm protocol $w WM_DELETE_WINDOW "set tclDevData(positions,goToDialog) [wm geometry $w];set goToDialog(ok) 0"

    set goToDialog(line) $lineNum

    set f [tixListNoteBook $w.f -dynamicgeometry 0]
    $f subwidget hlist add line -itemtype text -text "Line" -under 0
    $f subwidget hlist add procedure -itemtype text -text "Procedure" -under 0

    $f add line
    $f add procedure

    set p [$f subwidget line]
    set line [tixLabelEntry $p.line -label "Enter line number:" -labelside top -options {
        entry.textVariable goToDialog(line)
    }]
    bind [$p.line subwidget entry] <KeyRelease-Return> {set goToDialog(ok) 1}
    bind [$p.line subwidget entry] <KeyRelease-Escape> {set goToDialog(ok) 0}

    pack $line -side top -expand 1 -fill x -padx 3

    set p [$f subwidget procedure]
    tixLabelEntry $p.procedure -label "Enter procedure name:" -labelside top -options {
        entry.textVariable goToDialog(procedure)
    }
    pack $p.procedure -side top -expand 1 -fill x -padx 3
    bind [$p.procedure subwidget entry] <KeyRelease-Return> {set goToDialog(ok) 1}
    bind [$p.procedure subwidget entry] <KeyRelease-Escape> {set goToDialog(ok) 0}

    pack $f -side left -expand 1 -fill both

    set buttons [tixButtonBox $w.buttons -orientation vertical -pady 2 -relief flat]

    $buttons add goto -text "Go to" -command {set goToDialog(ok) 1}
    $buttons add cancel -text Cancel -command {set goToDialog(ok) 0}
    pack $buttons -fill y -expand 0 -side right -padx 3

    # run dialog
    #    focus [$f.findStr subwidget entry]
    tkwait visibility $w
    focus [$line subwidget entry]
    grab $w
    tkwait variable goToDialog(ok)
    grab release $w

    set what [$f raised]

    # store position
    set tclDevData(positions,goToDialog) [wm geometry $w]

    destroy $w

    if {$goToDialog(ok)} {
        if {$what == "line"} {
            set value $goToDialog(line)
        } else {
            set value $goToDialog(procedure)
        }

        return [list $what $value]
    } else {
        return {}
    }
}