##############################################################################
#    TCL Developer Studio
#
#    Copyright (C) 1999  Alexey Kakunin
#    small@star.spb.ru
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
##############################################################################

namespace eval options {
    namespace export runDialog setDefault read
    namespace export onTixOptions onUseHL onTabSize onFont
    
    variable OptionsDialog
    variable OptionsDialogData
    variable Options {tixScheme tixFontSet 
            useIndent useSintaxIndent indentSize 
            changeTabs tabSize useHL
            fontSize fontEncoding
            useTemplates useTemplatesForKeywords 
            useTemplatesForBrace useTemplatesForParen useTemplatesForBracket useTemplatesForQuoteDbl
            printCommand showLineNumbers afterLineNum defaultPrinter prettyOutput}
    
    # run options dialog
    proc runDialog {} {
        variable OptionsDialog
        global optionsDialog
        global tclDevData
        
        SetData
        
        CreateDialog
        
        tkwait visibility $OptionsDialog
        grab $OptionsDialog
        tkwait variable optionsDialog(ok)
        grab release $OptionsDialog
        
        set tclDevData(positions,optionsDialog) [wm geometry $OptionsDialog]
        
        destroy $OptionsDialog
        
        if {$optionsDialog(ok)} {
            # apply changes
            Apply
        }
        
        return
    }
    
    # create options dialog
    proc CreateDialog {} {
        variable OptionsDialog
        variable OptionsDialogData
        global optionsDialog
        global tclDevData
        
        set OptionsDialog [toplevel .optionsDialog]
        wm title $OptionsDialog "Options"
        
        if {[info exist tclDevData(positions,optionsDialog)] &&
             [string range $tclDevData(positions,optionsDialog) 0 2] != "1x1"} {
            wm geometry $OptionsDialog $tclDevData(positions,optionsDialog)
        } else {
            set x [expr {([winfo screenwidth .] - 100)/2}]
            set y [expr {([winfo screenheight .] - 100)/2}]
            
            wm geometry $OptionsDialog +$x+$y
        }
        
        wm protocol $OptionsDialog WM_DELETE_WINDOW "set tclDevData(positions,optionsDialog) [wm geometry $OptionsDialog];set optionsDialog(ok) 0"
        
        # create notebook
        set nb [tixNoteBook $OptionsDialog.nb]
        $nb subwidget nbframe config -font nbFrameFont -tabpadx 0 -tabpady 0
        pack $nb -side left -fill both -expand 1
        
        # add tabs
        set common [$nb add common -label "Common" -underline 0]
        set editor [$nb add editor -label "Editor" -underline 0]
        set font [$nb add font -label "Font" -underline 0]
        set print [$nb add print -label "Print" -underline 0]
        
        # create common tab:
        tixLabelFrame $common.tix -label "Tix Options" -padx 2 -pady 2
        set tixFrame [$common.tix subwidget frame]
        pack $common.tix -fill both -expand 1 -side top
        
        tixComboBox $tixFrame.scheme -label "Tix Scheme:" -editable false -options {
            entry.textVariable options::OptionsDialogData(tixScheme)
        }
        
        set lb [$tixFrame.scheme subwidget listbox]
        
        foreach name {Windows TK Gray Blue Bisque SGIGray TixGray} {
            $lb insert end $name
        }
        pack $tixFrame.scheme -side top -expand 1
        
        tixComboBox $tixFrame.fontSet -label "Tix FontSet:" -editable false -options {
            entry.textVariable options::OptionsDialogData(tixFontSet)
        }
        
        set lb [$tixFrame.fontSet subwidget listbox]
        
        foreach name {TK 12Point 14Point} {
            $lb insert end $name
        }
        pack $tixFrame.fontSet -side top -expand 1
        
        message $tixFrame.w -aspect 500 -text "Warning: this options will applied only for new windows"
        pack $tixFrame.w -side top -expand 1
        
        #create editor tab
        set f [frame $editor.f]
        pack $f -side top -expand 1 -fill both
        
        #insent options
        tixLabelFrame $f.indent -label "Indent" -padx 2 -pady 2
        pack $f.indent -side left -fill both -expand 1
        set indentFrame [$f.indent subwidget frame]
        checkbutton $indentFrame.useIndent -text "Use Indent" \
        -variable options::OptionsDialogData(useIndent)
        checkbutton $indentFrame.useSintaxIndent -text "Use Sintax Indent" \
        -variable options::OptionsDialogData(useSintaxIndent)
        tixControl $indentFrame.indentSize -label "Indent Size: " \
        -integer true -variable options::OptionsDialogData(indentSize) -min 1 -max 10 -options {
            entry.width 5
        }
        
        pack $indentFrame.useIndent $indentFrame.useSintaxIndent \
        -side top -expand 1 -anchor w
        
        pack $indentFrame.indentSize -side bottom -expand 1 -anchor w
        
        #tabs options
        tixLabelFrame $f.tabs -label "Tabs" -padx 2 -pady 2
        pack $f.tabs -side left -fill both -expand 1
        set tabsFrame [$f.tabs subwidget frame]
        checkbutton $tabsFrame.changeTabs -text "Change tabs by spaces" \
        -variable options::OptionsDialogData(changeTabs)
        tixControl $tabsFrame.tabSize -label "Tab Size: " \
        -integer true -variable options::OptionsDialogData(tabSize) -min 1 -max 10 -options {
            entry.width 5
        }
        
        pack $tabsFrame.changeTabs -side top -expand 1 -anchor w
        pack $tabsFrame.tabSize -side bottom -expand 1 -anchor w
        
        # highlighting options
        tixLabelFrame $editor.hl -label "Highlighting" -padx 2 -pady 2
        pack $editor.hl -side top -fill both -expand 1
        set hlFrame [$editor.hl subwidget frame]
        checkbutton $hlFrame.useHL -text "Use highlighting" \
        -variable options::OptionsDialogData(useHL)
        pack $hlFrame.useHL -side top -expand 1 -anchor w
        
        # templates options
        tixLabelFrame $editor.templ -label "Templates" -padx 2 -pady 2
        pack $editor.templ -side top -fill both -expand 1
        set templFrame [$editor.templ subwidget frame]
        checkbutton $templFrame.useTemp -text "Use Templates:" \
        -variable options::OptionsDialogData(useTemplates) -command options::OnUseTemplates
        pack $templFrame.useTemp -side top -anchor w
        
        set optionsDialog(forKeywords) [checkbutton $templFrame.forKeywords -text "for Keywords" \
        -variable options::OptionsDialogData(useTemplatesForKeywords) -command options::OnUseTemplates]
        pack $templFrame.forKeywords -side top -anchor w -padx 10
        
        frame $templFrame.f
        frame $templFrame.f.1
        frame $templFrame.f.2
        pack $templFrame.f -side top -expand 1 -fill both
        pack $templFrame.f.1 $templFrame.f.2 -side left -expand 1 -fill both
        
        set optionsDialog(forBrace) [checkbutton $templFrame.f.1.forBrace -text "for \{" \
        -variable options::OptionsDialogData(useTemplatesForBrace) -command options::OnUseTemplates]
        set optionsDialog(forParen) [checkbutton $templFrame.f.1.forParen -text "for \(" \
        -variable options::OptionsDialogData(useTemplatesForParen) -command options::OnUseTemplates]
        set optionsDialog(forBracket) [checkbutton $templFrame.f.2.forBracket -text "for \[" \
        -variable options::OptionsDialogData(useTemplatesForBracket) -command options::OnUseTemplates]
        set optionsDialog(forQuoteDbl) [checkbutton $templFrame.f.2.forQuoteDbl -text "for \"" \
        -variable options::OptionsDialogData(useTemplatesForQuoteDbl) -command options::OnUseTemplates]
        
        pack $templFrame.f.1.forBrace $templFrame.f.1.forParen -side top -anchor w -padx 10
        pack $templFrame.f.2.forBracket $templFrame.f.2.forQuoteDbl -side top -anchor w -padx 10
        
        # create font tabs
        tixControl $font.size -label "Size: " \
        -integer 0 -variable options::OptionsDialogData(fontSize) -min 6 -max 40 -options {
            label.width 10
            entry.width 5
        }
        tixLabelEntry $font.encoding -label "Encoding:" -options {
            label.width 10
            entry.textVariable options::OptionsDialogData(fontEncoding)
        }
        
        pack $font.size $font.encoding -side top -anchor w -pady 3 -padx 3
        
        #create printer tab
        tixLabelEntry $print.printCommand -label "Print command:" -options {
            label.width 20
            entry.textVariable options::OptionsDialogData(printCommand)
        }
        pack $print.printCommand -side top -fill x -anchor w -padx 3 -pady 2
        
        checkbutton $print.showLineNumbers -text "Show line numbers" \
  	        -variable options::OptionsDialogData(showLineNumbers) \
                -command "$print.f1.lineNum fconfigure -state \[lindex \{disabled normal\} \$options::OptionsDialogData(showLineNumbers)\]"
        pack $print.showLineNumbers -side top -anchor w -padx 3 -pady 1
        frame $print.f1
        pack $print.f1 -side top -fill x -anchor w -padx 3 -pady 2
        label $print.f1.l1 -text "    After each "
        entry $print.f1.lineNum -textvariable options::OptionsDialogData(afterLineNum) \
                                -state [lindex {disabled normal} $options::OptionsDialogData(showLineNumbers)]
        label $print.f1.l2 -text " lines"
        pack $print.f1.l1 $print.f1.lineNum $print.f1.l2 -side left -expand 1
        checkbutton $print.prettyOutput -text "Pretty Output" \
  	        -variable options::OptionsDialogData(prettyOutput)
        pack $print.prettyOutput -side top -anchor w -padx 3 -pady 2
        tixLabelEntry $print.defaultPrinter -label "Default Printer:" -options {
            label.width 20
            entry.textVariable options::OptionsDialogData(defaultPrinter)
        }        
        pack $print.defaultPrinter -side top -fill x -anchor w -padx 3 -pady 2
        
        set buttons [tixButtonBox $OptionsDialog.buttons -orientation vertical -pady 2 -relief flat]
        
        $buttons add ok -text "OK" -command {set optionsDialog(ok) 1}
        $buttons add apply -text "Apply" -command {options::Apply}
        $buttons add cancel -text Cancel -command {set optionsDialog(ok) 0}
        pack $buttons -fill y -expand 1 -side right -padx 3
        
        return
    }
    
    proc OnUseTemplates {} {
        global optionsDialog
        variable OptionsDialogData
        
        if {$OptionsDialogData(useTemplates)} {
            foreach  name {forBrace forParen forBracket forQuoteDbl}  {
                $optionsDialog($name) configure -state normal
            }
        } else  {
            foreach  name {forBrace forParen forBracket forQuoteDbl}  {
                $optionsDialog($name) configure -state disabled
            }
        }
        
        return
    }
    
    # set data from global array to the local dialog's array
    proc SetData {} {
        variable OptionsDialogData
        variable Options
        global tclDevData
        
        foreach name $Options {
            set OptionsDialogData($name) $tclDevData(options,$name) 
        }
        
        return
    }
    
    # set data from local dialog's array into the global array
    proc Apply {} {
        variable OptionsDialogData
        variable Options
        global tclDevData
        
        foreach name $Options {
            if {$tclDevData(options,$name) != $OptionsDialogData($name)} {
                set tclDevData(options,$name) $OptionsDialogData($name)
            }
        }
        
        return
    }
    
    # set default values into the global options array
    proc setDefault {} {
        global tcl_platform
        global tclDevData
        
        if {$tcl_platform(platform) == "windows"} {
            set tclDevData(options,fontSize) 14
            set tclDevData(options,fontEncoding) ""
        } else {
            set tclDevData(options,fontSize) 10
            set tclDevData(options,fontEncoding) "koi8-r"
        }
        
        font create nbFrameFont -size 8
        
        set tclDevData(options,tixScheme) TixGray
        set tclDevData(options,tixFontSet) 14Point
        
        set tclDevData(options,useIndent) 1
        set tclDevData(options,useSintaxIndent) 1
        set tclDevData(options,indentSize) 4
        
        set tclDevData(options,changeTabs) 1
        set tclDevData(options,tabSize) 4
        
        set tclDevData(options,useHL) 1
        
        set tclDevData(options,useTemplates) 1
        set tclDevData(options,useTemplatesForKeywords) 1
        set tclDevData(options,useTemplatesForBrace) 1
        set tclDevData(options,useTemplatesForParen) 1
        set tclDevData(options,useTemplatesForBracket) 1
        set tclDevData(options,useTemplatesForQuoteDbl) 1

        set tclDevData(options,printCommand) "a2ps"
        set tclDevData(options,showLineNumbers) 1
        set tclDevData(options,afterLineNum) 1
        set tclDevData(options,prettyOutput) "0"
        set tclDevData(options,defaultPrinter) "lp"
        
        return
    }
    
    proc CreateFonts {} {
        global tclDevData
        
        # set editor font
        set font "-*-courier-medium-r-*-*-"
        if {$tclDevData(options,fontSize) != ""} {
            append font $tclDevData(options,fontSize)
        } else {
            append font *
        }
        
        append font "-*-*-*-*-*-"
        
        if {$tclDevData(options,fontEncoding) != ""} {
            append font $tclDevData(options,fontEncoding)
        } else {
            append font *-*
        }
        
        set tclDevData(options,fonts,editorFont) $font
        
        # set comment font
        set font "-*-courier-medium-i-*-*-"
        if {$tclDevData(options,fontSize) != ""} {
            append font $tclDevData(options,fontSize)
        } else {
            append font *
        }
        
        append font "-*-*-*-*-*-"
        
        if {$tclDevData(options,fontEncoding) != ""} {
            append font $tclDevData(options,fontEncoding)
        } else {
            append font *-*
        }
        
        set tclDevData(options,fonts,commentFont) $font
        
        # set keyword font
        set font "-*-courier-bold-r-*-*-"
        if {$tclDevData(options,fontSize) != ""} {
            append font $tclDevData(options,fontSize)
        } else {
            append font *
        }
        
        append font "-*-*-*-*-*-"
        
        if {$tclDevData(options,fontEncoding) != ""} {
            append font $tclDevData(options,fontEncoding)
        } else {
            append font *-*
        }
        
        set tclDevData(options,fonts,keywordFont) $font
        
        return
    }
    
    proc read {} {
        global tclDevData
        global tcl_platform
        global TCLDEV_HOME
        global version
        
        set tclDevData(rcVersion) 1
        setDefault
        
        if {$tcl_platform(platform) =="windows"} {
            set rcName "$TCLDEV_HOME/tcldev.cfg"
        } else {
            set rcName "~/.tcldevrc"
        }
        
        catch {unset tclDevData(options,tclDevStudioVersion)}
        
        # check file
        if {[file exist $rcName]} {
            # load file
            catch {source $rcName}
        }
        
        #check version
        if {![info exist tclDevData(options,tclDevStudioVersion)] ||
            $tclDevData(options,tclDevStudioVersion) != $version} {
            #don't use dialog geometry
            foreach name [array names tclDevData positions,*] {
                unset tclDevData($name)
            }
        }
        
        return
    }
    
    proc save {} {
        global tclDevData
        global tcl_platform
        global TCLDEV_HOME
        global version
        
        if {$tcl_platform(platform) =="windows"} {
            set rcName "$TCLDEV_HOME/tcldev.cfg"
        } else {
            set rcName "~/.tcldevrc"
        }
        
        if {[catch {set fileID [open $rcName "w"]}]} {
            # can not open rc file
            return
        }
        
        
        puts $fileID "#This is TCL Developer Studio resource file"
        puts $fileID "set tclDevData(rcVersion) $tclDevData(rcVersion)"
        puts $fileID "set tclDevData(options,tclDevStudioVersion) $version"
        puts $fileID "# options"
        puts $fileID "set tclDevData(options,fontSize) $tclDevData(options,fontSize)"
        puts $fileID "set tclDevData(options,fontEncoding) $tclDevData(options,fontEncoding)"
        puts $fileID "set tclDevData(options,tixScheme) $tclDevData(options,tixScheme)"
        puts $fileID "set tclDevData(options,tixFontSet) $tclDevData(options,tixFontSet)"
        puts $fileID "set tclDevData(options,useIndent) $tclDevData(options,useIndent)"
        puts $fileID "set tclDevData(options,useSintaxIndent) $tclDevData(options,useSintaxIndent)"
        puts $fileID "set tclDevData(options,indentSize) $tclDevData(options,indentSize)"
        puts $fileID "set tclDevData(options,changeTabs) $tclDevData(options,changeTabs)"
        puts $fileID "set tclDevData(options,tabSize) $tclDevData(options,tabSize)"
        puts $fileID "set tclDevData(options,useHL) $tclDevData(options,useHL)"
        
        puts $fileID "# Window positions"
        puts $fileID "set tclDevData(positions,mainFrame) [mainFrame::getGeometry]"
        puts $fileID "set tclDevData(positions,projectWinSize) [mainFrame::getProjectWinSize]"
        puts $fileID "set tclDevData(positions,outputWinSize) [mainFrame::getOutputWinSize]"
        
        foreach dialogName {findDialog findInFilesDialog replaceDialog
            optionsDialog goToDialog projectDialog
            newDialog} {
            if {[info exist tclDevData(positions,$dialogName)]} {
                puts $fileID "set tclDevData(positions,$dialogName) $tclDevData(positions,$dialogName)"
            }
        }
        
        puts $fileID "# files information"
        
        if {[info exist tclDevData(recentFiles)]} {
            puts $fileID "set tclDevData(recentFiles) \{ $tclDevData(recentFiles) \}"
        }
        if {[info exist tclDevData(recentProjects)]} {        
            puts $fileID "set tclDevData(recentProjects) \{ $tclDevData(recentProjects) \}"
        }
        
        foreach name [array names tclDevData "files,*,cursor"] {
            if {$tclDevData($name) != ""} {
                puts $fileID "set tclDevData($name) $tclDevData($name)"
            }
        }
        foreach name [array names tclDevData "files,*,selFirst"] {
            if {$tclDevData($name) != ""} {
                puts $fileID "set tclDevData($name) $tclDevData($name)"
            }
        }
        foreach name [array names tclDevData "files,*,selLast"] {
            if {$tclDevData($name) != ""} {
                puts $fileID "set tclDevData($name) $tclDevData($name)"
            }
        }
        foreach name [array names tclDevData "files,*,anchor"] {
            if {$tclDevData($name) != ""} {
                puts $fileID "set tclDevData($name) $tclDevData($name)"
            }
        }
        
        
        close $fileID
    }
    
    #------------------------------------------------------
    # reactions on change options
    proc onTixOptions {name1 name2 op} {
        global tclDevData
        global tcl_platform
        
        if {[info exist tclDevData(options,tixScheme)]} {
            set scheme $tclDevData(options,tixScheme)
        } else {
            set scheme "Windows"
        }
        
        if {[info exist tclDevData(options,tixFontSet)]} {
            set fontSet $tclDevData(options,tixFontSet)
        } else {
            set fontSet "12Point"
        }
        
        if {$scheme == "Windows"} {
            tix resetoptions TixGray $fontSet
            
            global tixOption
            
            if {$tcl_platform(platform) == "windows"} {
                set tixOption(bg)           SystemMenu
            } else  {
                set tixOption(bg)           gray
            }
            
            set tixOption(fg)           black
            
            set tixOption(dark1_bg)     gray86
            set tixOption(dark1_fg)     black
            set tixOption(dark2_bg)     gray77
            set tixOption(dark2_fg)     black
            set tixOption(inactive_bg)  gray77
            set tixOption(inactive_fg)  black
            
            set tixOption(light1_bg)    white
            set tixOption(light1_fg)    white
            set tixOption(light2_bg)    white
            set tixOption(light2_fg)    white
            
            set tixOption(active_bg)    $tixOption(dark1_bg)
            set tixOption(active_fg)    $tixOption(fg)
            set tixOption(disabled_fg)  gray55
            
            set tixOption(input1_bg)    white
            set tixOption(input2_bg)    white
            set tixOption(output1_bg)   $tixOption(dark1_bg)
            set tixOption(output2_bg)   $tixOption(bg)
            
            set tixOption(select_fg)    black
            set tixOption(select_bg)    lightblue
            
            set tixOption(selector)    yellow
            
            option add *background         $tixOption(bg) 10
            option add *Background        $tixOption(bg) $tixOption(prioLevel)
            option add *background        $tixOption(bg) $tixOption(prioLevel)
            option add *Foreground        $tixOption(fg) $tixOption(prioLevel)
            option add *foreground        $tixOption(fg) $tixOption(prioLevel)
            option add *activeBackground    $tixOption(active_bg) $tixOption(prioLevel)
            option add *activeForeground    $tixOption(active_fg) $tixOption(prioLevel)
            option add *HighlightBackground    $tixOption(bg) $tixOption(prioLevel)
            option add *selectBackground    $tixOption(select_bg) $tixOption(prioLevel)
            option add *selectForeground    $tixOption(select_fg) $tixOption(prioLevel)
            option add *selectBorderWidth    0 $tixOption(prioLevel)
            option add *Menu.selectColor        $tixOption(selector) $tixOption(prioLevel)
            option add *TixMenu.selectColor        $tixOption(selector) $tixOption(prioLevel)
            option add *Menubutton.padY            5 $tixOption(prioLevel)
            option add *Button.borderWidth        2 $tixOption(prioLevel)
            option add *Button.anchor        c $tixOption(prioLevel)
            option add *Checkbutton.selectColor        $tixOption(selector) $tixOption(prioLevel)
            option add *Radiobutton.selectColor        $tixOption(selector) $tixOption(prioLevel)
            option add *Entry.relief        sunken $tixOption(prioLevel)
            option add *Entry.highlightBacground    $tixOption(bg) $tixOption(prioLevel)
            option add *Entry.background        $tixOption(input1_bg) $tixOption(prioLevel)
            option add *Entry.foreground        black $tixOption(prioLevel)
            option add *Entry.insertBackground    black $tixOption(prioLevel)
            option add *Label.anchor        w $tixOption(prioLevel)
            option add *Label.borderWidth        0 $tixOption(prioLevel)
            option add *Listbox.background        $tixOption(light1_bg) $tixOption(prioLevel)
            option add *Listbox.relief        sunken $tixOption(prioLevel)
            option add *Scale.foreground        $tixOption(fg) $tixOption(prioLevel)
            option add *Scale.activeForeground    $tixOption(bg) $tixOption(prioLevel)
            option add *Scale.background        $tixOption(bg) $tixOption(prioLevel)
            option add *Scale.sliderForeground    $tixOption(bg) $tixOption(prioLevel)
            option add *Scale.sliderBackground    $tixOption(light1_bg) $tixOption(prioLevel)
            option add *Scrollbar.background        $tixOption(bg) $tixOption(prioLevel)
            option add *Scrollbar.troughColor        $tixOption(light1_bg) $tixOption(prioLevel)
            option add *Scrollbar.relief        sunken $tixOption(prioLevel)
            option add *Scrollbar.borderWidth        1 $tixOption(prioLevel)
            option add *Scrollbar.width            15 $tixOption(prioLevel)
            option add *Text.background        $tixOption(input1_bg) $tixOption(prioLevel)
            option add *Text.relief        sunken $tixOption(prioLevel)
            option add *TixBalloon*background             #ffff60 $tixOption(prioLevel)
            option add *TixBalloon*foreground             black $tixOption(prioLevel)
            option add *TixBalloon.background             black $tixOption(prioLevel)
            option add *TixBalloon*Label.anchor             w $tixOption(prioLevel)
            option add *TixControl*entry.highlightBacground    $tixOption(bg) $tixOption(prioLevel)
            option add *TixControl*entry.background        $tixOption(input1_bg) $tixOption(prioLevel)
            option add *TixControl*entry.foreground        black $tixOption(prioLevel)
            option add *TixControl*entry.insertBackground    black $tixOption(prioLevel)
            option add *TixDirTree*Scrollbar.background        $tixOption(bg) $tixOption(prioLevel)
            option add *TixDirTree*Scrollbar.troughColor        $tixOption(light1_bg) $tixOption(prioLevel)
            option add *TixDirTree*hlist.highlightBacground    $tixOption(bg) $tixOption(prioLevel)
            option add *TixDirTree*hlist.background        $tixOption(light1_bg) $tixOption(prioLevel)
            option add *TixDirTree*hlist.activeBackground    $tixOption(light1_bg) $tixOption(prioLevel)
            option add *TixDirTree*hlist.disabledBackground    $tixOption(light1_bg) $tixOption(prioLevel)
            option add *TixDirTree*f1.borderWidth        1 $tixOption(prioLevel)
            option add *TixDirTree*f1.relief            sunken $tixOption(prioLevel)
            option add *TixDirList*Scrollbar.background        $tixOption(bg) $tixOption(prioLevel)
            option add *TixDirList*Scrollbar.troughColor        $tixOption(light1_bg) $tixOption(prioLevel)
            option add *TixDirList*hlist.highlightBacground    $tixOption(bg) $tixOption(prioLevel)
            option add *TixDirList*hlist.background        $tixOption(light1_bg) $tixOption(prioLevel)
            option add *TixDirList*hlist.activeBackground    $tixOption(light1_bg) $tixOption(prioLevel)
            option add *TixDirList*hlist.disabledBackground    $tixOption(light1_bg) $tixOption(prioLevel)
            option add *TixDirList*f1.borderWidth        1 $tixOption(prioLevel)
            option add *TixDirList*f1.relief            sunken $tixOption(prioLevel)
            option add *TixScrolledHList*Scrollbar.background        $tixOption(bg) $tixOption(prioLevel)
            option add *TixScrolledHList*Scrollbar.troughColor        $tixOption(light1_bg) $tixOption(prioLevel)
            option add *TixScrolledHList*hlist.highlightBacground    $tixOption(bg) $tixOption(prioLevel)
            option add *TixScrolledHList*hlist.background        $tixOption(light1_bg) $tixOption(prioLevel)
            option add *TixScrolledHList*hlist.activeBackground    $tixOption(light1_bg) $tixOption(prioLevel)
            option add *TixScrolledHList*hlist.disabledBackground    $tixOption(light1_bg) $tixOption(prioLevel)
            option add *TixScrolledHList*f1.borderWidth        1 $tixOption(prioLevel)
            option add *TixScrolledHList*f1.relief            sunken $tixOption(prioLevel)
            option add *TixTree*Scrollbar.background        $tixOption(bg) $tixOption(prioLevel)
            option add *TixTree*Scrollbar.troughColor        $tixOption(light1_bg) $tixOption(prioLevel)
            option add *TixTree*hlist.highlightBacground    $tixOption(bg) $tixOption(prioLevel)
            option add *TixTree*hlist.background        $tixOption(light1_bg) $tixOption(prioLevel)
            option add *TixTree*hlist.activeBackground    $tixOption(light1_bg) $tixOption(prioLevel)
            option add *TixTree*hlist.disabledBackground    $tixOption(light1_bg) $tixOption(prioLevel)
            option add *TixTree*f1.borderWidth        1 $tixOption(prioLevel)
            option add *TixTree*f1.relief            sunken $tixOption(prioLevel)
            option add *TixFileEntry*Entry.background         $tixOption(input1_bg) $tixOption(prioLevel)
            option add *TixHList.background            $tixOption(light1_bg) $tixOption(prioLevel)
            option add *TixHList.activeBackground        $tixOption(light1_bg) $tixOption(prioLevel)
            option add *TixHList.disabledBackground        $tixOption(light1_bg) $tixOption(prioLevel)
            option add *TixLabelEntry*entry.highlightBacground    $tixOption(bg) $tixOption(prioLevel)
            option add *TixLabelEntry*entry.background        $tixOption(input1_bg) $tixOption(prioLevel)
            option add *TixLabelEntry*entry.foreground        black $tixOption(prioLevel)
            option add *TixLabelEntry*entry.insertBackground    black $tixOption(prioLevel)
            option add *TixMultiList*Listbox.borderWidth        0 $tixOption(prioLevel)
            option add *TixMultiList*Listbox.highlightThickness    0 $tixOption(prioLevel)
            option add *TixMultiList*Scrollbar.background        $tixOption(bg) $tixOption(prioLevel)
            option add *TixMultiList*Scrollbar.troughColor        $tixOption(light1_bg) $tixOption(prioLevel)
            option add *TixMultiList*Scrollbar.relief        sunken $tixOption(prioLevel)
            option add *TixMultiList*Scrollbar.width        15 $tixOption(prioLevel)
            option add *TixMultiList*f1.borderWidth        2 $tixOption(prioLevel)
            option add *TixMultiList*f1.relief            sunken $tixOption(prioLevel)
            option add *TixMultiList*f1.highlightThickness        2 $tixOption(prioLevel)
            option add *TixMDIMenuBar*menubar.relief        raised $tixOption(prioLevel)
            option add *TixMDIMenuBar*menubar.borderWidth        2 $tixOption(prioLevel)
            option add *TixMDIMenuBar*Menubutton.padY         2 $tixOption(prioLevel)
            option add *TixNoteBook.Background            $tixOption(bg) $tixOption(prioLevel)
            option add *TixNoteBook.nbframe.Background        $tixOption(bg) $tixOption(prioLevel)
            option add *TixNoteBook.nbframe.backPageColor        $tixOption(bg) $tixOption(prioLevel)
            option add *TixNoteBook.nbframe.inactiveBackground    $tixOption(inactive_bg) $tixOption(prioLevel)
            option add *TixPanedWindow.handleActiveBg         $tixOption(active_bg) $tixOption(prioLevel)
            option add *TixPanedWindow.seperatorBg            $tixOption(bg) $tixOption(prioLevel)
            option add *TixPanedWindow.handleBg               $tixOption(dark1_bg) $tixOption(prioLevel)
            option add *TixPopupMenu*menubutton.background     $tixOption(dark1_bg) $tixOption(prioLevel)
            option add *TixScrolledHList*Scrollbar.background        $tixOption(bg) $tixOption(prioLevel)
            option add *TixScrolledHList*Scrollbar.troughColor        $tixOption(light1_bg) $tixOption(prioLevel)
            option add *TixScrolledHList*hlist.highlightBacground    $tixOption(bg) $tixOption(prioLevel)
            option add *TixScrolledHList*hlist.background        $tixOption(light1_bg) $tixOption(prioLevel)
            option add *TixScrolledTList*Scrollbar.background        $tixOption(bg) $tixOption(prioLevel)
            option add *TixScrolledTList*Scrollbar.troughColor        $tixOption(light1_bg) $tixOption(prioLevel)
            option add *TixScrolledTList*tlist.highlightBacground    $tixOption(bg) $tixOption(prioLevel)
            option add *TixScrolledTList*tlist.background        $tixOption(light1_bg) $tixOption(prioLevel)
            option add *TixScrolledListBox*Scrollbar.background        $tixOption(bg) $tixOption(prioLevel)
            option add *TixScrolledListBox*Scrollbar.troughColor        $tixOption(light1_bg) $tixOption(prioLevel)
            option add *TixScrolledListBox*listbox.highlightBacground    $tixOption(bg) $tixOption(prioLevel)
            option add *TixScrolledListBox*listbox.background        $tixOption(light1_bg) $tixOption(prioLevel)
            option add *TixScrolledText*Scrollbar.background        $tixOption(bg) $tixOption(prioLevel)
            option add *TixScrolledText*Scrollbar.troughColor        $tixOption(light1_bg) $tixOption(prioLevel)
            option add *TixScrolledWindow*Scrollbar.background        $tixOption(bg) $tixOption(prioLevel)
            option add *TixScrolledWindow*Scrollbar.troughColor        $tixOption(light1_bg) $tixOption(prioLevel)
            option add *TixScrolledWindow.frame.background        $tixOption(light1_bg) $tixOption(prioLevel)
            option add *TixTree*Scrollbar.background        $tixOption(bg) $tixOption(prioLevel)
            option add *TixTree*Scrollbar.troughColor        $tixOption(light1_bg) $tixOption(prioLevel)
            option add *TixTree*hlist.highlightBacground    $tixOption(bg) $tixOption(prioLevel)
            option add *TixTree*hlist.background        $tixOption(light1_bg) $tixOption(prioLevel)
            option add *TixTree*hlist.borderWidth        1 $tixOption(prioLevel)
            option add *TixComboBox*Entry.highlightBacground        $tixOption(bg) $tixOption(prioLevel)
            option add *TixComboBox*Entry.background            $tixOption(input1_bg) $tixOption(prioLevel)
            option add *TixComboBox*Entry.foreground            black $tixOption(prioLevel)
            option add *TixComboBox*Entry.insertBackground        black $tixOption(prioLevel)
        } else  {
            tix resetoptions $scheme $fontSet
        }
    }
    
    proc onUseHL {name1 name2 op} {
        global tclDevData
        
        if {$tclDevData(options,useHL)} {
            editorWindows::enableHL
        } else {
            editorWindows::disableHL
        }
    }
    
    proc onTabSize {name1 name2 op} {
        editorWindows::onTabSize
        return
    }
    
    proc onFont {name1 name2 op} {
        global tclDevData
        
        if {![info exist tclDevData(options,fontSize)] ||
            ![info exist tclDevData(options,fontEncoding)]} {
            return
        }
        
        CreateFonts
        
        editorWindows::onFontChange
    }
    
    #------------------------------------------------------    
    # create reactions on change options
    trace variable tclDevData(options,tixScheme) w options::onTixOptions
    trace variable tclDevData(options,tixFontSet) w options::onTixOptions
    trace variable tclDevData(options,useHL) w options::onUseHL
    trace variable tclDevData(options,tabSize) w options::onTabSize
    trace variable tclDevData(options,fontSize) w options::onFont
    trace variable tclDevData(options,fontEncoding) w options::onFont
}