##############################################################################
#    TCL Developer Studio
#
#    Copyright (C) 1999  Alexey Kakunin
#    small@star.spb.ru
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
##############################################################################

#All information about project is stored in tclDevData(project,...)
namespace eval project {
    
    variable vars {Name Dir RunCmd DebugCmd FileComment}
    variable Platforms {both unix windows}
    
    namespace export loadFile save resetProject
    namespace export runDialog
    
    #data for dialog
    variable OptionsDialog
    variable OptionsDialogData
    
    # Reset all project variables
    proc resetProject {} {
        variable vars
        variable Platforms
        global tclDevData
        
        foreach varName $vars {
            foreach platform $Platforms {
                set tclDevData(project,$platform,$varName) ""
            }
        }
        
        return
    }
    
    # load project file
    proc loadFile { fileName } {
        variable vars
        variable Platfroms
        global tclDevData
        global TCLDEV_HOME
        
        resetProject
        
        if {[catch {set fileID [open $fileName "r"]}]} {
            resetProject
            error "Can not load project $fileName"
        }
        
        #create temp-file name
        set tempDir $TCLDEV_HOME
        
        set index 0
        set tmpName "$TCLDEV_HOME/tmpprj.$index"
        
        while {[file exist $tmpName]} {
            incr index
            set tmpName "$TCLDEV_HOME/tmpprj.$index"
        }
        
        #open temp file
        if {[catch {set tmpFileID [open $tmpName "w"]}]} {
            resetProject
            close $fileID
            error "Can not open tmp-file"
        }
        
        set line ""
        
        #find start of project settings
        while {[gets $fileID line] != -1} {
            if {$line == "##TCLDevStudioProject"} {
                break
            }
        }
        
        if {$line != "##TCLDevStudioProject"} {
            close $fileID
            close $tmpFileID
            file delete -force $tmpName
            resetProject
            error "$fileName is not TCL Developer Studio project!"
        }
        
        #copy all lines to temp file
        while {[gets $fileID line] != -1} {
            #is it end of properties?
            if {$line == "##TCLDevStudioProject_END"} {
                break
            }
            #is it comment?
            if {[string index $line 0] != "#"} {
                #this is not comment
                break
            }
            
            #copy all string except first character into the tmp-file
            puts $tmpFileID [string range $line 1 end]
        }
        
        close $fileID
        close $tmpFileID
        
        if {$line != "##TCLDevStudioProject_END"} {
            file delete -force $tmpName
            resetProject
            error "$fileName is not TCL Developer Studio project!"
        }
        
        if {[catch "source $tmpName"]} {
            file delete -force $tmpName
            resetProject
            error "$fileName is not TCL Developer Studio project!"
        }
        
        file delete -force $tmpName
        
        # change current directory to the Project Directory
        if {[getDir] != ""} {
            catch {cd [getDir]}
        }
        
        return
    }
    
    # save project settings
    proc save {fileID} {
        variable vars
        variable Platforms
        global tclDevData
        
        #save TCLDev Comment
        puts $fileID "##TCLDevStudioProject"
        #save attention comment
        puts $fileID "## Attention! Followed data is TCL Developer Studio Project information."
        puts $fileID "## All data from it place to TCLDevStudioProject_END will removed and rewritted during edition in TCLDevStudio."
        puts $fileID "## Do not edit it yourself!"
        puts $fileID "#global tclDevData"
        puts $fileID "#set tclDevData(project,version) 2"
        
        # for each variable
        foreach varName $vars {
            #for each platform
            foreach platform $Platforms {
                #save variable
                if {[info exist tclDevData(project,$platform,$varName)]} {
                    puts $fileID "#set tclDevData(project,$platform,$varName) \"[FormatString $tclDevData(project,$platform,$varName)]\""
                }
            }
        }
        
        #write and of Project Information
        puts $fileID "##TCLDevStudioProject_END"
        
        return         
    }
    
    # create accessors
    global tclDevData
    global tcl_platform
    
    foreach varName $vars {
        set procName get
        append procName $varName
        
        set name "tclDevData(project,both,$varName)"
        set pName "tclDevData(project,$tcl_platform(platform),$varName)"
        
        if {$varName != "FileComment"} {
            proc $procName {} "global tclDevData;        \
            if {$$pName != \"\"} {    \
                return $$pName };   \
            return $$name"
        }
        
        namespace export $procName    
    }
    
    #for FileComment define special accessor
    proc getFileComment { fileName } {
        global tclDevData
        global tcl_platform
        
        if {$tclDevData(project,$tcl_platform(platform),FileComment) != ""} {
            set text [ConvertToText $tclDevData(project,$tcl_platform(platform),FileComment)]
        } else {
            set text [ConvertToText $tclDevData(project,both,FileComment)]
        }
        
        return [clock format [clock seconds] -format $text]
    }
    namespace export getFileComment
    
    #reset project
    resetProject
    
    # set reaction on change project settings
    trace variable tclDevData(project,both,Name) w mainFrame::setProjectName
    foreach platform $Platforms {
        trace variable tclDevData(project,$platform,Name) w mainFrame::setProjectName
    }
    
    #------------------------------------------------------
    # project dialog implementation
    proc runDialog {} {
        variable ProjectDialog
        global projectDialog
        global tclDevData
        
        SetData
        
        CreateDialog
        
#        grab $ProjectDialog
        tkwait variable projectDialog(ok)
#        grab release $ProjectDialog
        
        set tclDevData(positions,projectDialog) [wm geometry $ProjectDialog]
        
        if {$projectDialog(ok)} {
            # apply changes
            Apply
        }
        
        destroy $ProjectDialog
        
        return
    }
    
    # set data from project into the local project dialog data    
    proc SetData {} {
        variable vars
        variable Platforms
        variable ProjectDialogData
        global tclDevData
        
        #for each variable
        foreach name $vars {
            #for each platform
            foreach platform $Platforms {
                set pVarName $name
                append pVarName "_"
                append pVarName $platform
                
                if {[info exist tclDevData(project,$platform,$name)]} {
                    if {$name == "Dir" || $name == "RunCmd" || $name == "DebugCmd"} {
                        set ProjectDialogData($platform,$name) [ConvertPathToView $tclDevData(project,$platform,$name)]
                    } else  {
                        set ProjectDialogData($platform,$name) $tclDevData(project,$platform,$name)     
                    }
                } else {
                    set ProjectDialogData($platform,$name) ""
                }
            }
        }
        
        return
    }
    
    proc Apply {} {
        variable vars
        variable Platforms
        variable ProjectDialogData
        global tclDevData
        global projectDialog
        
        #for each platform
        foreach platform $Platforms {
            #read data from dialog
            set ProjectDialogData($platform,FileComment) \
            [ConvertFromText [[$projectDialog($platform,fileComment) subwidget text] get 1.0 "end -1c"]]
            
            #for each variable
            foreach name $vars {
                if {$name == "Dir" || $name == "RunCmd" || $name == "DebugCmd"} {
                    set tclDevData(project,$platform,$name) [ConvertPathFromView $ProjectDialogData($platform,$name)]
                } else  {
                    set tclDevData(project,$platform,$name) $ProjectDialogData($platform,$name)     
                }
            }
        }
        
        return
    }
    
    proc CreateDialog {} {
        variable ProjectDialog
        variable ProjectDialogData
        variable Platforms
        global projectDialog
        global tclDevData
        
        set ProjectDialog [toplevel .optionsDialog]
        wm title $ProjectDialog "Project Settings"
        
        if {[info exist tclDevData(positions,projectDialog)] &&
             [string range $tclDevData(positions,projectDialog) 0 2] != "1x1"} {
            wm geometry $ProjectDialog $tclDevData(positions,projectDialog)
        } else {
            set x [expr {([winfo screenwidth .] - 100)/2}]
            set y [expr {([winfo screenheight .] - 100)/2}]
            
            wm geometry $ProjectDialog 530x210+$x+$y
        }
        
        wm protocol $ProjectDialog WM_DELETE_WINDOW "set tclDevData(positions,projectDialog) [wm geometry $ProjectDialog];set projectDialog(ok) 0"
        
        #create listbook
        set lnb [tixListNoteBook $ProjectDialog.lnb -dynamicgeometry 0]
        pack $lnb -fill both -expand 1 -padx 1 -pady 1 -side left
        
        #fill list tabs
        foreach name $Platforms {
            $lnb subwidget hlist add $name -itemtype text -text "$name" -under 0
            
            set this [$lnb add $name]
            
            # create notebooks
            set nb [tixNoteBook $this.nb]
            $nb subwidget nbframe config -font nbFrameFont -tabpadx 0 -tabpady 0
            pack $nb -side left -fill both -expand 1
            
            # add tabs
            set general [$nb add common -label "General" -underline 0]
            set commands [$nb add commands -label "Commands" -underline 0]
            set comments [$nb add comments -label "Comments" -underline 0]
            
            #create general tab
            eval "tixLabelEntry $general.name -label \"Name:\" -options {label.width 14 entry.textVariable project::ProjectDialogData($name,Name)}"
            eval "tixFileEntry $general.dir -label \"Working Dir:\" -dialogtype tixDirSelectDialog -options {label.width 14 entry.textVariable project::ProjectDialogData($name,Dir)}"
            
            pack $general.name $general.dir -side top -padx 3 -pady 3 -anchor w
            
            #create commands tab
            eval "tixFileEntry $commands.run -label \"Run:\" -dialogtype tixExFileSelectDialog -options {label.width 14 entry.textVariable project::ProjectDialogData($name,RunCmd) }"
            eval "tixFileEntry $commands.debug -label \"Debug:\" -dialogtype tixExFileSelectDialog -options {label.width 14 entry.textVariable project::ProjectDialogData($name,DebugCmd) }"
            
            pack $commands.run $commands.debug -side top -padx 3 -pady 3 -anchor w
            
            #create comments tab
            label $comments.fileLabel -text "File comment:"
            pack $comments.fileLabel -side top -anchor w
            
            set projectDialog($name,fileComment) [tixScrolledText $comments.fileComment -options {
                text.width 20
                text.wrap none}]
            pack $comments.fileComment -side top -fill both -expand 1
            [$comments.fileComment subwidget text] insert 1.0 [ConvertToText $ProjectDialogData($name,FileComment)]
        }
        
        set buttons [tixButtonBox $ProjectDialog.buttons -orientation vertical -pady 2 -relief flat]
        
        $buttons add ok -text "OK" -command {set projectDialog(ok) 1}
        $buttons add apply -text "Apply" -command {project::Apply}
        $buttons add cancel -text Cancel -command {set projectDialog(ok) 0}
        pack $buttons -fill y -expand 1 -side right -padx 3
        
        return
    }
    
    #convert from text widget. Change all EOL to \n
    proc ConvertFromText {text} {
        set l $text
        set result ""
        set newLine "\\n"
        
        while {[set index [string first "\n" $l]] != -1} {
            # copy from start string to index
            set result $result[string range $l 0 [expr {$index - 1}]]$newLine
            # copy end of string
            incr index
            set l [string range $l $index end]
        }
        # copy last end
        set result $result$l
        
        return $result
    }
    
    #convert to text widget. Change all \n to EOL
    proc ConvertToText {text} {
        
        set l $text        
        # convert \\n -> \n
        set result ""
        set newLine "\n"
        
        while {[set index [string first "\\n" $l]] != -1} {
            # copy from start string to index
            append result [string range $l 0 [expr {$index - 1}]]
            append result $newLine
            
            # copy end of string
            incr index
            incr index
            
            set l [string range $l $index end]
        }
        
        # copy last end
        set result $result$l
        
        return $result
    }
    
}